#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that can perform a resolution transformation on a
molecule.
"""
from collections import defaultdict
from itertools import product, combinations

import networkx as nx

from ..molecule import Molecule, attributes_match
from .processor import Processor
from ..utils import are_all_equal, format_atom_string
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


def build_graph_mapping_collection(from_ff, to_ff, mappings):
    """
    Function that produces a collection of :class:`vermouth.map_parser.Mapping`
    objects.
    Hereby deprecated.

    Parameters
    ----------
    from_ff: vermouth.forcefield.ForceField
        Origin force field.
    to_ff: vermouth.forcefield.ForceField
        Destination force field.
    mappings: dict[str, dict[str, vermouth.map_parser.Mapping]]
        All known mappings

    Returns
    -------
    collections.abc.Iterable
        A collection of mappings that map from `from_ff` to `to_ff`.
    """
    return mappings[from_ff.name][to_ff.name].values()


def edge_matcher(graph1, graph2, node11, node12, node21, node22):
    """
    Checks whether the resids for node11 and node12 in graph1 are the same, and
    whether that's also true for node21 and node22 in graph2.

    Parameters
    ----------
    graph1: networkx.Graph
    graph2: networkx.Graph
    node11: collections.abc.Hashable
        A node key in `graph1`.
    node12: collections.abc.Hashable
        A node key in `graph1`.
    node21: collections.abc.Hashable
        A node key in `graph2`.
    node22: collections.abc.Hashable
        A node key in `graph2`.

    Returns
    -------
    bool
    """
    node11 = graph1.nodes[node11]
    node12 = graph1.nodes[node12]
    node21 = graph2.nodes[node21]
    node22 = graph2.nodes[node22]
    return (node11.get('resid') == node12.get('resid')) ==\
           (node21.get('resid') == node22.get('resid'))


def node_matcher(node1, node2):
    """
    Checks whether nodes should be considered equal for isomorphism. Takes all
    attributes in `node2` into account, except for the attributes "atype",
    "charge", "charge_group", "resid", "replace", and "_old_atomname".

    Parameters
    ----------
    node1: dict
    node2: dict

    Returns
    -------
    bool
    """
    return attributes_match(node1, node2,
                            ignore_keys=('atype', 'charge', 'charge_group',
                                         'resid', 'replace', '_old_atomname'))


def _old_atomname_match(node1, node2):
    """
    Adds a _name attribute to copies of the nodes, and feeds it all to
    :func:`node_matcher`
    """
    name1 = node1.get('_old_atomname', node1['atomname'])
    name2 = node2.get('_old_atomname', node2['atomname'])
    node1 = node1.copy()
    node2 = node2.copy()
    node1['_name'] = name1
    node2['_name'] = name2
    del node1['atomname']
    del node2['atomname']
    return node_matcher(node1, node2)


def node_should_exist(modification, node_idx):
    """
    Returns True if the node with index `node_idx` in `modification` should
    already exist in the parent molecule.

    Parameters
    ----------
    modification: networkx.Graph
    node_idx: collections.abc.Hashable
        The key of a node in `modification`.

    Returns
    -------
    bool
        True iff the node `node_idx` in `modification` should already exist in
        the parent molecule.
    """
    return not modification.nodes[node_idx].get('PTM_atom', False)


def ptm_resname_match(mol_node, map_node):
    """
    As :func:`node_matcher`, except that empty resname and false PTM_atom
    attributes from `node2` are removed.
    """
    if 'resname' in map_node and not map_node['resname']:
        map_node = map_node.copy()
        del map_node['resname']
    if 'PTM_atom' in map_node and not map_node['PTM_atom']:
        map_node = map_node.copy()
        del map_node['PTM_atom']
    if 'modifications' in mol_node:
        map_node = map_node.copy()
        matching_mod = all(map_mod in mol_node.get('modifications', [])
                           for map_mod in map_node.pop('modifications', []))
    else:
        matching_mod = True
    is_equal = node_matcher(mol_node, map_node)
    return is_equal and matching_mod


def cover(to_cover, options):
    """
    Implements a recursive backtracking algorithm to cover all elements of
    `to_cover` with the elements from `options` that have the lowest index.
    In this context "to cover" means that all items in an element of `options`
    must be in `to_cover`. Elements in `to_cover` can only be covered *once*.

    Parameters
    ----------
    to_cover: collections.abc.MutableSet
        The items that should be covered.
    options: collections.abc.Sequence[collections.abc.MutableSet]
        The elements that can be used to cover `to_cover`. All items in an
        element of `options` must be present in `to_cover` to qualify.

    Returns
    -------
    None or list
        None if no covering can be found, or the list of items from `options`
        with the lowest indices that exactly covers `to_cover`.
    """
    if not to_cover:
        return []
    for idx, option in enumerate(options):
        if all(item in to_cover for item in option):
            left_to_cover = to_cover.copy()
            for item in option:
                # Only remove the leftmost item. PS. we know for sure all items
                # in option are in left_to_cover at least once.
                left_to_cover.remove(item)
            found = cover(left_to_cover, options[idx:])
            if found is not None:
                return [option] + found
    return None


def get_mod_mappings(mappings):
    """
    Returns a dict of all known modification mappings.

    Parameters
    ----------
    mappings: collections.abc.Iterable[vermouth.map_parser.Mapping]
        All known mappings.

    Returns
    -------
    dict[tuple[str], vermouth.map_parser.Mapping]
        All mappings that describe a modification mapping.
    """
    out = {}
    for mapping in mappings:
        if mapping.type == 'modification':
            out[mapping.names] = mapping
    return out


def modification_matches(molecule, mappings):
    """
    Returns a minimal combination of modification mappings and where they
    should be applied that describes all modifications in `molecule`.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule whose modifications should be treated. Modifications are
        described by the 'modifications' node attribute.
    mappings: collections.abc.Iterable[vermouth.map_parser.Mapping]
        All known mappings.

    Returns
    -------
    list[tuple[dict, vermouth.molecule.Link, dict]]
        A list with the following items:
            Dict describing the correspondence of node keys in `molecule` to
                node keys in the modification.

            The modification.

            Dict with all reference atoms, mapping modification nodes to
                nodes in `molecule`.
    """
    modified_nodes = set()  # This will contain whole residues.
    for idx, node in molecule.nodes.items():
        if node.get('modifications', []):
            modified_nodes.add(idx)
    ptm_subgraph = molecule.subgraph(modified_nodes)
    grouped = nx.connected_components(ptm_subgraph)
    found_ptm_groups = []
    # For every modification group we would like a set with the names of the
    # involved modifications, so we can use that to figure out which mod
    # mappings should be used.
    for group in grouped:
        modifications = {
            mod.name for mol_idx in group for mod in molecule.nodes[mol_idx].get('modifications', [])
        }
        found_ptm_groups.append(modifications)
    needed_mod_mappings = set()
    known_mod_mappings = get_mod_mappings(mappings)
    for group in found_ptm_groups:
        # known_mod_mappings is a dict[tuple[str], Mapping]. We want to know
        # the minimal combination of those needed to cover all the PTMs found
        # in group. The cheapest solution is covering the names of the PTMs in
        # group with keys from known_mod_mappings. An improvement would be to
        # do the graph covering again.
        # TODO?
        covered_by = cover(list(group),
                           sorted(known_mod_mappings, key=len, reverse=True))
        if covered_by is None:
            LOGGER.warning("Can't find modification mappings for the "
                           "modifications {}. The following modification "
                           "mappings are known: {}",
                           list(group), known_mod_mappings)
            continue
        needed_mod_mappings.update(covered_by)
    matches = []
    # Sort on the tuple[str] type names of the mappings so that mappings that
    # define most modifications at the same time get processed first
    for mod_name in sorted(needed_mod_mappings, key=len, reverse=True):
        mod_mapping = known_mod_mappings[mod_name]
        for mol_to_mod, modification, references in mod_mapping.map(molecule, node_match=ptm_resname_match):
            matches.append((mol_to_mod, modification, references))
            if not set(mol_to_mod) <= modified_nodes:
                # TODO: better message
                LOGGER.warning('Overlapping modification mappings')
            modified_nodes -= set(mol_to_mod)
    return matches


def apply_block_mapping(match, molecule, graph_out, mol_to_out, out_to_mol):
    """
    Performs a mapping operation for a "block". `match` is a tuple of 3
    elements that describes what nodes in `molecule` should correspond to
    a :class:`vermouth.molecule.Block` that should be added to `graph_out`, and
    any atoms that should be used a references.
    Add the required :class:`vermouth.molecule.Block` to `graph_out`, and
    updates `mol_to_out` and `out_to_mol` *in-place*.

    Parameters
    ----------
    match
    molecule: networkx.Graph
        The original molecule
    graph_out: vermouth.molecule.Molecule
        The newly created graph that describes `molecule` at a different
        resolution.
    mol_to_out: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        A dict mapping nodes in `molecule` to nodes in `graph_out` with the
        associated weights.
    out_to_mol: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        A dict mapping nodes in `graph_out` to nodes in `molecule` with the
        associated weights.

    Returns
    -------
    set
        A set of all overlapping nodes that were already mapped before.
    set
        A set of none-to-one mappings. I.e. nodes that were created without
        nodes mapping to them.
    dict
        A dict of reference atoms, mapping `graph_out` nodes to nodes in
        `molecule`.
    """
    mol_to_block, blocks_to, references = match
    if graph_out.nrexcl is None:
        graph_out.nrexcl = blocks_to.nrexcl
    try:
        # merge_molecule will return a dict mapping the node keys of the
        # added block to the ones in graph_out
        # FIXME: Issue #154 lives here.
        block_to_out = graph_out.merge_molecule(blocks_to)
    except ValueError:
        # This probably means the nrexcl of the block is different from the
        # others. This means the user messed up their data. Or there are
        # different forcefields in the same forcefield folder...
        LOGGER.exception('Residue(s) {} is not compatible with the others',
                         set(nx.get_node_attributes(blocks_to, 'resname').values()),
                         type='inconsistent-data')
        raise
    # overlap does not have to be a dict, since the values in block_to_out are
    # guaranteed to be unique in graph_out. So we can look them up in
    # mol_to_out
    overlap = set(mol_to_out.keys()) & set(mol_to_block.keys())
    for mol_idx in mol_to_block:
        for block_idx, weight in mol_to_block[mol_idx].items():
            out_idx = block_to_out[block_idx]
            mol_to_out[mol_idx][out_idx] = weight
            out_to_mol[out_idx][mol_idx] = weight

    none_to_one_mappings = set()
    mapped_block_idxs = {block_idx
                         for mol_idx in mol_to_block
                         for block_idx in mol_to_block[mol_idx]}
    for spawned in set(blocks_to.nodes) - mapped_block_idxs:
        # These nodes come from "nowhere", so, let's pretend they come from
        # all nodes in the block. This helps with setting attributes such
        # as 'chain'
        # "None to one" mapping - this is fine. This happens with e.g.
        # charge dummies.
        spawned = block_to_out[spawned]
        none_to_one_mappings.add(spawned)
        for mol_idx in mol_to_block:
            mol_to_out[mol_idx][spawned] = 0
            out_to_mol[spawned][mol_idx] = 0

    new_references = {block_to_out[mod_idx]: mol_idx for mod_idx, mol_idx in references.items()}
    return overlap, none_to_one_mappings, new_references


def apply_mod_mapping(match, molecule, graph_out, mol_to_out, out_to_mol):
    """
    Performs the mapping operation for a modification.

    Parameters
    ----------
    match
    molecule: networkx.Graph
        The original molecule
    graph_out: vermouth.molecule.Molecule
        The newly created graph that describes `molecule` at a different
        resolution.
    mol_to_out: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        A dict mapping nodes in `molecule` to nodes in `graph_out` with the
        associated weights.
    out_to_mol: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        A dict mapping nodes in `graph_out` to nodes in `molecule` with the
        associated weights.

    Returns
    -------
    dict[str, dict[tuple, vermouth.molecule.Link]]
        A dict of all modifications that have been applied by this modification
        mapping operations. Maps interaction type to involved atoms to the
        modification responsible.
    dict
        A dict of reference atoms, mapping `graph_out` nodes to nodes in
        `molecule`.
    """
    mol_to_mod, modification, references = match
    LOGGER.info('Applying modification mapping {}', modification.name, type='general')
    mod_to_mol = defaultdict(dict)
    for mol_idx, mod_idxs in mol_to_mod.items():
        for mod_idx in mod_idxs:
            mod_to_mol[mod_idx][mol_idx] = mol_to_mod[mol_idx][mod_idx]
    mod_to_mol = dict(mod_to_mol)
    mod_to_out = {}
    # Some nodes of modification will already exist. The question is
    # which, and which index they have in graph_out.
    for mod_idx in modification:
        if not node_should_exist(modification, mod_idx):
            # Node does not exist yet.
            if not graph_out.nodes:
                out_idx = 0
            else:
                out_idx = max(graph_out) + 1
            mod_to_out[mod_idx] = out_idx
            graph_out.add_node(out_idx, **modification.nodes[mod_idx])
        else:
            # Node should already exist
            # We need to find the out_index of this node. Since the
            # node already exists, there is at least one mol_idx in
            # mol_to_out that refers to the correct out_idx. What we do
            # is try to find those mol indices by looking at
            # mod_to_mol.
            # Find the other mol nodes that map to this bead according
            # to the mod mapping...
            mol_idxs = mod_to_mol[mod_idx]
            # ...and make the node with the correct attributes.
            out_idxs = set()
            for mol_idx in mol_idxs:
                out_idxs.update(mol_to_out.get(mol_idx, {}))
            for out_idx in out_idxs:
                out_node = graph_out.nodes[out_idx]
                if modification.nodes[mod_idx]['atomname'] == out_node['atomname']:
                    break
            else:  # No break, so no matching node found
                raise ValueError("No node found in molecule with "
                                 "atomname {}".format(modification.nodes[mod_idx]['atomname']))
            # Undefined loop variable is guarded against by the else-raise above
            mod_to_out[mod_idx] = out_idx  # pylint: disable=undefined-loop-variable
            graph_out.nodes[out_idx].update(modification.nodes[mod_idx].get('replace', {})) # pylint: disable=undefined-loop-variable
        graph_out.nodes[out_idx]['modifications'] = graph_out.nodes[out_idx].get('modifications', [])
        if modification not in graph_out.nodes[out_idx]['modifications']:
            graph_out.nodes[out_idx]['modifications'].append(modification)

    for mol_idx in mol_to_mod:
        for mod_idx, weight in mol_to_mod[mol_idx].items():
            out_idx = mod_to_out[mod_idx]
            if mol_idx not in mol_to_out:
                mol_to_out[mol_idx] = {}
            mol_to_out[mol_idx][out_idx] = weight
            if out_idx not in out_to_mol:
                out_to_mol[out_idx] = {}
            out_to_mol[out_idx][mol_idx] = weight

    for mod_idx, mod_jdx in modification.edges:
        out_idx = mod_to_out[mod_idx]
        out_jdx = mod_to_out[mod_jdx]
        if not graph_out.has_edge(out_idx, out_jdx):
            graph_out.add_edge(out_idx, out_jdx)

    new_references = {mod_to_out[mod_idx]: mol_idx for mod_idx, mol_idx in references.items()}

    # Apply interactions
    applied_interactions = defaultdict(lambda: defaultdict(list))
    for interaction_type, interactions in modification.interactions.items():
        for interaction in interactions:
            atoms = [mod_to_out[mod_idx] for mod_idx in interaction.atoms]
            assert len(atoms) == len(interaction.atoms)
            interaction = interaction._replace(atoms=atoms)
            applied_interactions[interaction_type][tuple(atoms)].append(modification)
            graph_out.add_interaction(interaction_type, **interaction._asdict())
    return dict(applied_interactions), new_references


def attrs_from_node(node, attrs):
    """
    Helper function that applies a "replace" operations on the node if
    required, and then returns a dict of the attributes listed in `attrs`.

    Parameters
    ----------
    node: dict
    attrs: collections.abc.Container
        Attributes that should be in the output.

    Returns
    -------
    dict
    """
    if 'replace' in node:
        node = node.copy()
        node.update(node['replace'])
    return {attr: val for attr, val in node.items() if attr in attrs}


def do_mapping(molecule, mappings, to_ff, attribute_keep=(), attribute_must=()):
    """
    Creates a new :class:`~vermouth.molecule.Molecule` in force field `to_ff`
    from `molecule`, based on `mappings`. It does this by doing a subgraph
    isomorphism of all blocks in `mappings` and `molecule`. Will issue warnings
    if there's atoms not contibuting to the new molecule, or if there's
    overlapping blocks.
    Node attributes in the new molecule will come from the blocks constructing
    it, except for those in `attribute_keep`, which lists the attributes that
    will be kept from `molecule`.

    Parameters
    ----------
    molecule: :class:`~vermouth.molecule.Molecule`
        The molecule to transform.
    mappings: dict[str, dict[str, dict[str, tuple]]]
        ``{ff_name: {ff_name: {block_name: (mapping, weights, extra)}}}``
        A collection of mappings, as returned by e.g.
        :func:`~vermouth.map_input.read_mapping_directory`.
    to_ff: :class:`~vermouth.forcefield.ForceField`
        The force field to transform to.
    attribute_keep: :class:`~collections.abc.Iterable`
        The attributes that will always be transferred from `molecule` to the
        produced graph.
    attribute_must: :class:`~collections.abc.Iterable`
        The attributes that the nodes in the output graph *must* have. If
        they're not provided by the mappings/blocks they're taken from
        `molecule`.

    Returns
    -------
    :class:`~vermouth.molecule.Molecule`
        A new molecule, created by transforming `molecule` to `to_ff` according
        to `mappings`.
    """
    attribute_keep = tuple(attribute_keep)
    attribute_must = tuple(attribute_must)
    # Transfering the meta maybe should be a copy, or a deep copy...
    # If it breaks we look at this line.
    graph_out = Molecule(force_field=to_ff, meta=molecule.meta)
    mappings = build_graph_mapping_collection(molecule.force_field, to_ff, mappings)
    block_matches = []
    for mapping in mappings:
        if mapping.type == 'block':
            block_matches.extend(mapping.map(molecule, node_match=_old_atomname_match,
                                             edge_match=edge_matcher))
    mod_matches = modification_matches(molecule, mappings)

    # Sort by lowest node key per residue. We need to do this, since
    # merge_molecule creates new resid's in order.
    block_sort_key = lambda x: min(x[0].keys())
    # Sort modifications by the lowest mapped index when all touched atoms are
    # newly created PTM atoms that do not exist yet. Otherwise, take the
    # highest index. If we don't do this we run the risk that a PTM mapped to
    # a node with a low index also changes a node of a higher residue, causing
    # all sorts of havok.
    mod_sort_key = lambda x: (max(x[0].keys())
                              if any(node_should_exist(x[1], idx)
                                     for idx in
                                     {mod_idx for mol_idx in x[0]
                                      for mod_idx in x[0][mol_idx]})
                              else
                              min(x[0].keys()))
    block_matches = sorted(block_matches, key=block_sort_key, reverse=True)
    mod_matches = sorted(mod_matches, key=mod_sort_key, reverse=True)

    # There are a few separate mapping cases to be considered:
    # One to one mapping - e.g. AA to AA, the simplest case
    # Many to one mapping - e.g. AA to CG without sharing atoms between beads
    # Many to many mapping - e.g. AA to CG *with* sharing atoms between bonds
    # These three cases are covered by the normal operation, the following are
    # caught with some additional logic
    # None to one - whole block taken as origin, with weights 0
    # One to none - unmapped atoms (produces a warning)
    # One to many - e.g. CG to AA. This mostly works, but we don't know how to
    #               make sure the "many" should be connected togeher. Gives a
    #               warning if it's disconnected.

    mol_to_out = defaultdict(dict)
    out_to_mol = defaultdict(dict)
    overlapping_mappings = set()
    none_to_one_mappings = set()
    modified_interactions = {}
    all_references = {}
    all_matches = []
    while block_matches or mod_matches:
        # Take the match with the lowest atom id, and prefer blocks over
        # modifications
        if (not block_matches or
                (mod_matches and
                 mod_sort_key(mod_matches[-1]) < block_sort_key(block_matches[-1]))):
            match = mod_matches.pop(-1)
            applied_interactions, refs = apply_mod_mapping(match,
                                                           molecule, graph_out,
                                                           mol_to_out, out_to_mol)
            modified_interactions.update(applied_interactions)
        else:
            match = block_matches.pop(-1)
            overlap, none_to_one, refs = apply_block_mapping(match,
                                                             molecule, graph_out,
                                                             mol_to_out, out_to_mol)
            overlapping_mappings.update(overlap)
            none_to_one_mappings.update(none_to_one)
        all_matches.append(match)
        all_references.update(refs)

    # At this point, we should have created graph_out at the desired
    # resolution, *and* have the associated correspondence in mol_to_out and
    # out_to_mol.

    # Set node attributes based on what the original atoms are.
    to_remove = set()
    for out_idx in out_to_mol:
        mol_idxs = out_to_mol[out_idx].keys()
        # Keep track of what bead comes from where
        subgraph = molecule.subgraph(mol_idxs)
        graph_out.nodes[out_idx]['graph'] = subgraph
        weights = out_to_mol[out_idx]
        graph_out.nodes[out_idx]['mapping_weights'] = weights

        if out_idx in all_references:
            ref_idx = all_references[out_idx]
            new_attrs = attrs_from_node(molecule.nodes[ref_idx],
                                        attribute_keep+attribute_must)
            for attr, val in new_attrs.items():
                # Attrs in attribute_keep we always transfer, those in
                # attribute_must we transfer only if they're not already in the
                # created node
                if attr in attribute_keep or attr not in graph_out.nodes[out_idx]:
                    graph_out.nodes[out_idx].update(new_attrs)
        else:
            attrs = defaultdict(list)
            for mol_idx in mol_idxs:
                new_attrs = attrs_from_node(molecule.nodes[mol_idx],
                                            attribute_keep+attribute_must)
                for attr, val in new_attrs.items():
                    attrs[attr].append(val)
            attrs_not_sane = []
            for attr, vals in attrs.items():
                if attr in attribute_keep or attr not in graph_out.nodes[out_idx]:
                    if vals:
                        graph_out.nodes[out_idx][attr] = vals[0]
                    else:
                        # No nodes hat the attribute.
                        graph_out.nodes[out_idx][attr] = None
                if not are_all_equal(vals):
                    attrs_not_sane.append(attr)
            if attrs_not_sane:
                LOGGER.warning('The attributes {} for atom {} are going to'
                               ' be garbage.', attrs_not_sane, format_atom_string(graph_out.nodes[out_idx]),
                               type='inconsistent-data')
        if graph_out.nodes[out_idx].get('atomname', '') is None:
            to_remove.add(out_idx)

    # We need to add edges between residues. Within residues comes from the
    # blocks.
    for match1, match2 in combinations(all_matches, 2):
        match1 = match1[0]
        match2 = match2[0]
        edges = molecule.edges_between(match1.keys(), match2.keys())
        # TODO: Backmapping needs love here
        for mol_idx, mol_jdx in edges:
            # Substract none_to_one_mappings, since those should not be made to
            # connect to things automatically.
            out_idxs = mol_to_out[mol_idx].keys() - none_to_one_mappings
            out_jdxs = mol_to_out[mol_jdx].keys() - none_to_one_mappings
            for out_idx, out_jdx in product(out_idxs, out_jdxs):
                if out_idx != out_jdx:
                    graph_out.add_edge(out_idx, out_jdx)

    ############################
    # Sanity check the results #
    ############################

    # "Many to one" mapping - overlapping blocks means dubious node properties
    if overlapping_mappings:
        LOGGER.warning('These atoms are covered by multiple blocks. This is a '
                       'bad idea: {}. This probably means the following output'
                       ' particles are wrong: {}.',
                       {format_atom_string(molecule.nodes[mol_idx])
                        for mol_idx in overlapping_mappings},
                       {format_atom_string(graph_out.nodes[out_idx], atomid='')
                        for mol_idx in overlapping_mappings
                        for out_idx in mol_to_out[mol_idx]},
                       type='inconsistent-data')

    # "One to many" mapping - not necessarrily a problem, unless it leads to
    # missing edges
    for mol_idx in mol_to_out:
        # Substract the none to one mapped nodes, since those don't contribute
        # and make false positives.
        out_idxs = mol_to_out[mol_idx].keys() - none_to_one_mappings
        if len(out_idxs) > 1 and not nx.is_connected(graph_out.subgraph(out_idxs)):
            # In this case there's a single input particle mapping to multiple
            # output particles. This probably means there's bonds missing
            LOGGER.warning('The input particle {} maps to multiple output '
                           'particles: {}, which are disconnected. There are '
                           'probably edges missing.',
                           format_atom_string(molecule.nodes[mol_idx]),
                           {format_atom_string(graph_out.nodes[out_idx], atomid='')
                            for out_idx in out_idxs},
                           type='inconsistent-data')

    # "One to none" mapping - this means your mapping files are incomplete
    uncovered_atoms = set(molecule.nodes.keys()) - set(mol_to_out.keys())
    if uncovered_atoms:
        uncovered_hydrogens = {idx for idx in uncovered_atoms
                               if molecule.nodes[idx].get('element', '') == 'H'}
        if uncovered_hydrogens:
            # Maybe this should be info?
            LOGGER.debug('These hydrogen atoms are not covered by a mapping.'
                         ' This is not the best idea. {}',
                         [format_atom_string(molecule.nodes[idx])
                          for idx in uncovered_hydrogens],
                         type='unmapped-atom'
                         )
        other_uncovered = uncovered_atoms - uncovered_hydrogens
        if other_uncovered:
            LOGGER.warning("These atoms are not covered by a mapping. Either"
                           " your mappings don't describe all atoms (bad idea),"
                           " or, there's no mapping available for all residues."
                           " {}",
                           [format_atom_string(molecule.nodes[idx])
                            for idx in other_uncovered],
                           type='unmapped-atom')

    for interaction_type in modified_interactions:
        for atoms, modifications in modified_interactions[interaction_type].items():
            if len(modifications) != 1:
                # TODO: better message
                LOGGER.warning('Interaction set by multiple modification '
                               'mappings', type='inconsistent-data')

    graph_out.remove_nodes_from(to_remove)
    return graph_out


class DoMapping(Processor):
    def __init__(self, mappings, to_ff, delete_unknown=False, attribute_keep=(),
                 attribute_must=()):
        self.mappings = mappings
        self.to_ff = to_ff
        self.delete_unknown = delete_unknown
        self.attribute_keep = tuple(attribute_keep)
        self.attribute_must = tuple(attribute_must)
        super().__init__()

    def run_molecule(self, molecule):
        return do_mapping(
            molecule,
            mappings=self.mappings,
            to_ff=self.to_ff,
            attribute_keep=self.attribute_keep,
            attribute_must=self.attribute_must
        )

    def run_system(self, system):
        mols = []
        for molecule in system.molecules:
            try:
                new_molecule = self.run_molecule(molecule)
            except KeyError as err:
                if not self.delete_unknown:
                    raise err
                else:
                    raise
                    # TODO: raise a loud warning here
            else:
                if new_molecule:
                    mols.append(new_molecule)
        system.molecules = mols
        system.force_field = self.to_ff



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handle the ITP file format from Gromacs.
"""

import copy
import itertools

__all__ = ['write_molecule_itp', ]


def _attr_has_not_none_attr(obj, attr):
    """
    Raise a Value error is 'obj' does not have an attribute 'attr' or if its
    value is ``None``.
    """
    try:
        value = getattr(obj, attr)
    except AttributeError:
        value = None
    if value is None:
        raise ValueError('{} has no attribute "{}".'.format(obj, attr))


def _interaction_sorting_key(interaction):
    ifdef = interaction.meta.get('ifdef')
    ifndef = interaction.meta.get('ifndef')
    if ifdef is not None and ifndef is not None:
        raise ValueError('An interaction cannot have both an "ifdef" '
                         'and an "ifndef" meta attribute.')
    if ifdef is not None:
        conditional = (ifdef, True)
    elif ifndef is not None:
        conditional = (ifndef, False)
    else:
        conditional = ()

    group = interaction.meta.get('group')
    if group is None:
        group = ''

    return (conditional, group)


def write_molecule_itp(molecule, outfile, header=(), moltype=None,
                       post_section_lines=None, pre_section_lines=None):
    """
    Write a molecule in ITP format.

    The molecule must have a `nrexcl` attribute. Each atom in the molecule must
    have at least the following keys: `atype`, `resid`, `resname`, `atomname`,
    and `charge_group`. Atoms can also have a `charge` and a `mass` key.

    If the `moltype` argument is not provided, then the molecule must have a
    "moltype" meta attribute.

    Parameters
    ----------
    molecule: Molecule
        The molecule to write. See above for the minimal information the
        molecule must contain.
    outfile: io.TextIOBase
        The file in which to write.
    header: collections.abc.Iterable[str]
        List of lines to write as comment at the beginning of the file. The
        comment character and the new line should not be included as they will
        be added in the function.
    moltype: str, optional
        The molecule type. If set to `None` (default), the molecule type is
        read from the "moltype" key of `molecule.meta`.
    post_section_lines: dict[str, collections.abc.Iterable[str]], optional
        List of lines to write at the end of some sections of the file. The
        argument is passed as a dict with the keys being the name of the
        sections, and the values being the lists of lines. If the argument is
        set to `None`, the lines will be read from the "post_section_lines" key
        of `molecule.meta`.
    pre_section_lines: dict[str, collections.abc.Iterable[str]], optional
        List of lines to write at the beginning of some sections, just after
        the section header. The argument is formatted in the same way as
        `post_section_lines`. If the argument is set to `None`, the lines will
        be read from the "post_section_lines" key of `molecule.meta`.

    Raises
    ------
    ValueError
        The molecule is missing required information.
    """
    # Make sure the molecule contains the information required to write the
    # header.
    if moltype is None:
        moltype = molecule.meta.get('moltype')
        if  moltype is None:
            raise ValueError('A molecule must have a moltype to write an '
                             'ITP, provide it with the moltype argument, or '
                             'with the "moltype" meta attribute of the molecule.')
    _attr_has_not_none_attr(molecule, 'nrexcl')

    # Make sure the molecule contains the information required to write the
    # [atoms] section. The charge and mass can be ommited, if so gromacs take
    # them from the [atomtypes] section of the ITP file.
    for attribute in ('atype', 'resid', 'resname', 'atomname',
                      'charge_group'):
        if not all([attribute in atom for _, atom in molecule.atoms]):
            raise ValueError('Not all atom have a {}.'.format(attribute))

    # Get the maximum length of each atom field so we can align the fields.
    # Atom indexes are written as a consecutive series starting from 1.
    # The maximum index of a 0-based series is `len(x) - 1`; because the
    # series starts at 1, the maximum value is `len(x).
    max_length = {'idx': len(str(len(molecule)))}
    for attribute in ('atype', 'resid', 'resname', 'atomname',
                      'charge_group', 'charge', 'mass'):
        max_length[attribute] = max(len(str(atom.get(attribute, '')))
                                    for _, atom in molecule.atoms)

    # Write the header.
    # We want to follow the header with an empty line, only if there is a
    # header. The `has_header` variable is needed in case `header` is a
    # generator, in which case we cannot know before hand if it contains lines.
    has_header = False
    for line in header:
        outfile.write('; {}\n'.format(line))
        has_header = True
    if has_header:
        outfile.write('\n')

    outfile.write('[ moleculetype ]\n')
    outfile.write('{} {}\n\n'.format(moltype, molecule.nrexcl))

    # Get the post- and pre- section lines. These lines are will be written at
    # the end or at the beginning of the relevant sections.
    if post_section_lines is None:
        post_section_lines = molecule.meta.get('post_section_lines', {})
    if pre_section_lines is None:
        pre_section_lines = molecule.meta.get('pre_section_lines', {})
    seen_sections = set()

    # The atoms in the [atoms] section must be consecutively numbered, yet
    # there is no guarantee that the molecule fulfill that constrain.
    # Therefore we renumber the atoms. The `correspondence` dict allows to
    # keep track of the correspondence between the original and the new
    # numbering so we can apply the renumbering to the interactions.
    # The resid and charge_group should also be consecutive, though this is
    # left as the user responsibility. Make sure residues and charge groups are
    # correctly numbered.
    correspondence = {}
    outfile.write('[ atoms ]\n')
    seen_sections.add('atoms')
    for line in pre_section_lines.get('atoms', []):
        outfile.write(line + '\n')
    for idx, (original_idx, atom) in enumerate(molecule.atoms, start=1):
        correspondence[original_idx] = idx
        new_atom = copy.copy(atom)
        # The charge and the mass can be blank and read from the [atomtypes]
        # section of the ITP file.
        new_atom['charge'] = new_atom.get('charge', '')
        new_atom['mass'] = new_atom.get('mass', '')

        outfile.write('{idx:>{max_length[idx]}} '
                      '{atype:<{max_length[atype]}} '
                      '{resid:>{max_length[resid]}} '
                      '{resname:<{max_length[resname]}} '
                      '{atomname:<{max_length[atomname]}} '
                      '{charge_group:>{max_length[charge_group]}} '
                      '{charge:>{max_length[charge]}} '
                      '{mass:>{max_length[mass]}}\n'
                      .format(idx=idx, max_length=max_length, **new_atom))
    for line in post_section_lines.get('atoms', []):
        outfile.write(line + '\n')
    outfile.write('\n')

    # Write the interactions
    conditional_keys = {True: '#ifdef', False: '#ifndef'}
    for name in molecule.sort_interactions(molecule.interactions):
        interactions = molecule.interactions[name]

        # Improper dihedral angles have their own section in the Molecule
        # object to distinguish them from the proper dihedrals. Yet, they
        # should be written under the [ dihedrals ] section of the ITP file.
        if name == 'impropers':
            name = 'dihedrals'
        outfile.write('[ {} ]\n'.format(name))
        seen_sections.add(name)
        for line in pre_section_lines.get(name, []):
            outfile.write(line + '\n')
        interactions_group_sorted = sorted(
            interactions,
            key=_interaction_sorting_key
        )
        interaction_grouped = itertools.groupby(
            interactions_group_sorted,
            key=_interaction_sorting_key
        )
        for (conditional, group), interactions_in_group in interaction_grouped:
            if conditional:
                conditional_key = conditional_keys[conditional[1]]
                outfile.write('{} {}\n'.format(conditional_key, conditional[0]))
            if group:
                outfile.write('; {}\n'.format(group))
            for interaction in interactions_in_group:
                atoms = ['{atom_idx:>{max_length[idx]}}'
                         .format(atom_idx=correspondence[x],
                                 max_length=max_length)
                         for x in interaction.atoms]
                parameters = ' '.join(str(x) for x in interaction.parameters)
                comment = ''
                if 'comment' in interaction.meta:
                    comment = ' ; ' + interaction.meta['comment']
                if name == 'virtual_sitesn':
                    to_join = [atoms[0], parameters] + atoms[1:]
                else:
                    to_join = atoms + [parameters]
                outfile.write(' '.join(to_join) + comment + '\n')
            if conditional:
                outfile.write('#endif\n')
            for line in post_section_lines.get(name, []):
                outfile.write(line + '\n')
            outfile.write('\n')

    # Some sections may have pre or post lines, but no other content. I that
    # case, we need to write the sections separately.
    remaining_sections = set(pre_section_lines) | set(post_section_lines)
    remaining_sections -= seen_sections
    for name in remaining_sections:
        outfile.write('[ {} ]\n'.format(name))
        for line in pre_section_lines.get(name, []):
            outfile.write(line + '\n')
        for line in post_section_lines.get(name, []):
            outfile.write(line + '\n')
        outfile.write('\n')



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a processor to assign molecule type names to molecules.

A molecule type (moltype) is Gromacs's concept of a molecule. Providing a name
for a molecule type is required to write an ITP file for that molecule. We also
use the molecule type name to group molecules sharing the same molecule type.
Molecule type identity is tested based on
:meth:`vermouth.molecule.Molecule.share_moltype_with`.
"""

from .processor import Processor


class NameMolType(Processor):
    """
    Assigns molecule type (moltype) names to molecules.

    Moltype names are the names given to molecules in an ITP file. This
    processor assign consecutive names to the molecule. If the `deduplicate`
    argument is set to `True`, then the processor assigns the same name to all
    molecules with the same topology.

    By default, the moltype name is written under the "moltype" key of the
    molecule meta attributes. This key can be changed with the `meta_key`
    argument.

    Parameters
    ----------
    deduplicate: bool
        If `True`, the same name is given to all the molecules that share the
        same topology. Else, each molecule is given a different name.
    meta_key: str
        The name of the key in the molecule `meta` dictionary under which the
        moltype must be stored.

    See Also
    --------
    vermouth.processors.set_molecule_meta.SetMoleculeMeta
        This processor can set key/value pairs in the meta attributes of one
        molecule, or all molecules in a system. It can be used to set the
        moltype manually.
    vermouth.gmx.itp.write_molecule_itp
        Writes the ITP file for a molecule, and use the 'moltype' meta to name
        the molecule.
    """
    # TODO: See issue #35
    def __init__(self, deduplicate=True, meta_key='moltype'):
        self.deduplicate = deduplicate
        self.meta_key = meta_key
        super().__init__()

    def run_system(self, system):
        if not system.molecules:
            return system

        if self.deduplicate:
            self._name_with_deduplication(system)
        else:
            self._name_without_deduplication(system)

        return system

    def _name_with_deduplication(self, system):
        group_id = 0
        representatives = [(group_id, system.molecules[0])]
        for molecule in system.molecules:
            for match_id, template in representatives:
                if molecule.share_moltype_with(template):
                    break
            else:  # no break
                group_id += 1
                representatives.append((group_id, molecule))
                match_id = group_id
            molecule.meta[self.meta_key] = 'molecule_{}'.format(match_id)

    def _name_without_deduplication(self, system):
        for molecule_id, molecule in enumerate(system.molecules):
            molecule.meta[self.meta_key] = 'molecule_{}'.format(molecule_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads quotes, and produces a random one.
"""

import os.path
import random

from ..file_writer import open
from .processor import Processor
from .. import DATA_PATH
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))

QUOTE_FILE = os.path.join(DATA_PATH, 'quotes.txt')


def read_quote_file(filehandle):
    """
    Iterates over `filehandle`, and yields all strings that are not empty.

    Parameters
    ----------
    filehandle: collections.abc.Iterable[str]
        A file opened for reading.

    Yields
    ------
    str
        All stripped elements of `filehandle` that are not empty.
    """
    for line in filehandle:
        line = line.strip()
        if line:
            yield line


class Quoter(Processor):
    """
    Processor that can produce random string taken from a file. Useful for e.g.
    quotes.

    Parameters
    ----------
    quote_file: pathlib.Path or str
        The path of the file containing the strings. Must contain at least one
        line.
    """
    def __init__(self, quote_file=None):
        if quote_file is None:
            quote_file = QUOTE_FILE
        self._quote_file = quote_file

    def run_system(self, system):
        """
        Logs a random line from the file passed at initialization.

        Parameters
        ----------
        system
            Not used

        Returns
        -------
        None
        """
        with open(self._quote_file) as handle:
            all_quotes = list(read_quote_file(handle))
        LOGGER.info(random.choice(all_quotes))



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that generates positions for nodes based on the weighted
average of the positions of the atoms they are constructed from.
"""


import numpy as np
from .processor import Processor


def do_average_bead(molecule, ignore_missing_graphs=False, weight=None):
    """
    Set the position of the particles to the mean of the underlying atoms.

    This requires the atoms to have a 'graph' attributes. By default, a
    :exc:`ValueError` is raised if any atom in the molecule is missing that
    'graph' attribute. This behavior can be changed by setting the
    'ignore_missing_graphs' argument to `True`, then the average positions are
    computed, but the atoms without a 'graph' attribute are skipped.

    The average is weighted using the 'mapping_weights' atom attribute. If the
    'mapping_weights' attribute is set, it has to be a dictionary with the
    atomname from the underlying graph as keys, and the weights as values.
    Atoms without a weight set use a default weight of 1.

    The average can also be weighted using an arbitrary node attribute by
    giving the attribute name with the `weight` keyword argument. This can be
    used to get the center of mass for instance; assuming the mass of the
    underlying atoms is stored under the "mass" attribute, setting `weight` to
    "mass" will place the bead at the center of mass. By default, `weight` is
    set to `None` and the center of geometry is used.

    The atoms in the underlying graph must have a position. If they do not,
    they are ignored from the average.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to update. The attribute `position` of the particles
        is updated on place. The nodes of the molecule must have an attribute
        `graph` that contains the subgraph of the initial molecule.
    ignore_missing_graphs: bool
        If `True`, skip the atoms that do not have a `graph` attribute; else
        fail if not all the atoms in the molecule have a `graph` attribute.
    weight: collections.abc.Hashable
        The name of the attribute used to weight the position of the node. The
        attribute is read from the underlying atoms.
    """
    # Make sure the molecule fullfill the requirements.
    missing = []
    for node in molecule.nodes.values():
        if 'graph' not in node:
            missing.append(node)
        elif weight is not None:
            have_all_weights = all(
                weight in subnode for subnode in node['graph'].nodes.values()
            )
            if not have_all_weights:
                raise KeyError('Not all underlying atoms have an attribute {}.'
                               .format(weight))
    if missing and not ignore_missing_graphs:
        raise ValueError('{} particles are missing the graph attribute'
                         .format(len(missing)))

    for node in molecule.nodes.values():
        if 'graph' in node:
            positions = np.array([
                subnode['position']
                for subnode in node['graph'].nodes().values()
                if subnode.get('position') is not None
            ])
            weights = np.array([
                node.get('mapping_weights', {}).get(subnode_key, 1) * subnode.get(weight, 1)
                for subnode_key, subnode in node['graph'].nodes.items()
                if subnode.get('position') is not None
            ])
            try:
                ndim = positions.shape[1]
            except IndexError:
                ndim = 3
            if abs(sum(weights)) < 1e-7:
                node['position'] = np.array([np.nan]*ndim, dtype=float)
            else:
                node['position'] = np.average(positions, axis=0, weights=weights)

    return molecule


class DoAverageBead(Processor):
    def __init__(self, ignore_missing_graphs=False, weight=None):
        super().__init__()
        self.ignore_missing_graphs = ignore_missing_graphs
        self.weight = weight

    def run_molecule(self, molecule):
        if self.weight is None:
            weight = molecule.force_field.variables.get('center_weight', None)
        elif self.weight is False:
            weight = None
        else:
            weight = self.weight
        return do_average_bead(molecule, self.ignore_missing_graphs, weight=weight)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide some helper classes to allow new style brace formatting for logging and
processing the `type` keyword.
"""
from collections import defaultdict
import logging


class Message:
    """
    Class that defers string formatting until it's ``__str__`` method is
    called.
    """
    def __init__(self, fmt, args, kwargs):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.fmt.format(*self.args, **self.kwargs)

    def __repr__(self):
        return '"{}".format(*{}, **{})'.format(self.fmt, self.args, self.kwargs)


class PassingLoggerAdapter(logging.LoggerAdapter):
    """
    Helper class that is actually capable of chaining multiple LoggerAdapters.
    """
    def __init__(self, logger, extra=None):
        if extra is None:
            extra = {}
        # These are all set by the logger.setter property. Which
        # super().__init__ calls.
        super().__init__(logger, extra)
        # A LoggerAdapter does not have a manager, but logging.Logger.log needs
        # it to see if logging is enabled.

    @property
    def manager(self):
        """
        .. autoattribute:: logging.Logger.manager
        """
        return self.logger.manager

    @manager.setter
    def manager(self, new_value):
        self.logger.manager = new_value

    def process(self, msg, kwargs):
        # The documentation is a lie and the original implementation clobbers
        # 'extra' that is set by other LoggerAdapters in the chain.
        # LoggerAdapter's process method is FUBARed, and aliases kwargs and
        # self.extra. And that's all it does. So we do it here by hand to make
        # sure we actually have an 'extra' attribute.
        # It should maybe be noted that generally this method gets executed
        # multiple times, so occasionally self.extra items are very persistent.
        if 'extra' not in kwargs:
            kwargs['extra'] = {}
        kwargs['extra'].update(self.extra)
        try:
            # logging.Logger does not have a process.
            msg, kwargs = self.logger.process(msg, kwargs)
        except AttributeError:
            pass
        return msg, kwargs

    def log(self, level, msg, *args, **kwargs):
        # Differs from super().log because this calls `self.logger.log` instead
        # of self.logger._log. LoggerAdapters don't have a _log.
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            if isinstance(self.logger, logging.Logger):
                # logging.Logger.log throws a hissy fit if it gets too many
                # kwargs, so leave just the ones known.
                kwargs = {key: val for key, val in kwargs.items()
                          if key in ['level', 'msg', 'exc_info', 'stack_info', 'extra']}
                self.logger._log(level, msg, args, **kwargs)  # pylint: disable=protected-access
            else:
                self.logger.log(level, msg, *args, **kwargs)

    def addHandler(self, *args, **kwargs):  # pylint: disable=invalid-name
        self.logger.addHandler(*args, **kwargs)


class StyleAdapter(PassingLoggerAdapter):
    """
    Logging adapter that encapsulate messages in :class:`Message`, allowing
    ``{}`` style formatting.
    """
    def log(self, level, msg, *args, **kwargs):
        # We need a different `log` method, since `Message` needs the args
        # as well as the kwargs. Otherwise it could've been done in process.
        # You can probably work around that by giving Message a __mod__ method,
        # but that's too much effort for now.
        msg, kwargs = self.process(msg, kwargs)
        super().log(level, Message(msg, args, kwargs), **kwargs)


class TypeAdapter(PassingLoggerAdapter):
    """
    Logging adapter that takes the `type` keyword argument passed to logging
    calls and passes adds it to the `extra` attribute.

    Parameters
    ----------
    logger: logging.Logger or logging.LoggerAdapter
        As described in :class:`logging.LoggerAdapter`.
    extra: dict
        As described in :class:`logging.LoggerAdapter`.
    default_type: str
        The type of the messages if none is given.
    """
    def __init__(self, logger, extra=None, default_type='general'):
        super().__init__(logger, extra)
        self.default_type = default_type

    def process(self, msg, kwargs):
        msg, kwargs = super().process(msg, kwargs)
        type_ = kwargs.pop('type', self.default_type)
        if 'type' not in kwargs['extra']:
            kwargs['extra']['type'] = type_
        return msg, kwargs


class BipolarFormatter:  # pylint: disable=too-few-public-methods
    """
    A logging formatter that formats using either `low_formatter` or
    `high_formatter` depending on the `logger`'s effective loglevel.

    Parameters
    ----------
    low_formatter: logging.Formatter
        The formatter used if `cutoff` <= `logger.getEffectiveLevel()`.
    high_formatter: logging.Formatter
        The formatter used if `cutoff` > `logger.getEffectiveLevel()`.
    cutoff: int
        The cutoff used to decide whether the low or high formatter is used.
    logger: logging.Logger
        The logger whose effective loglevel is used. Defaults to
        ``logging.getLogger()``.
    """
    def __init__(self, low_formatter, high_formatter, cutoff, logger=None):
        self.detailed_formatter = low_formatter
        self.pretty_formatter = high_formatter
        self.cutoff = cutoff
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

    def __getattr__(self, item):
        # Delegate to pretty formatter if the loglevel is high enough, and the
        # detailed formatter otherwise
        loglevel = self.logger.getEffectiveLevel()
        if loglevel > self.cutoff:
            return getattr(self.pretty_formatter, item)
        else:
            return getattr(self.detailed_formatter, item)


class CountingHandler(logging.NullHandler):
    """
    A logging handler that counts the number of times a specific type of
    message is logged per loglevel.

    Parameters
    ----------
    type_attribute: str
        The name of the attribute carrying the type.
    default_type: str
        The type of message if none is provided.

    """
    def __init__(self, *args, type_attribute='type', default_type='general', **kwargs):
        super().__init__(*args, **kwargs)
        self.counts = defaultdict(lambda: defaultdict(int))
        self.default_type = default_type
        self.type_attr = type_attribute

    def handle(self, record):
        """
        Handle a log record by counting it.
        """
        record_level = record.levelno
        record_type = getattr(record, self.type_attr, self.default_type)
        self.counts[record_level][record_type] += 1

    def number_of_counts_by(self, level=None, type=None):
        """
        Return the number of logging calls counted, filtered by level and type.

        Parameters
        ----------
        level
            Only count log events of this level.
        type
            Only count log events of this type.

        Returns
        -------
        int
            The number of events counted.
        """
        out = 0
        for lvl, type_counts in self.counts.items():
            if level is not None and level != lvl:
                continue
            for type_, count in type_counts.items():
                if type is not None and type != type_:
                    continue
                out += count
        return out


def get_logger(name):
    """
    Convenience method that wraps a :class:`TypeAdapter` around
    ``logging.getLogger(name)``

    Parameters
    ----------
    name: str
        The name of the logger to get. Passed to :func:`logging.getLogger`.
        Should probably be ``__name__``.
    """
    return TypeAdapter(logging.getLogger(name))


def ignore_warnings_and_count(counter, specifications, level=logging.WARNING):
    """
    Count the warnings after deducting the ones to ignore.

    Warnings to ignore are specified as tuple ``(<warning-type>, <count>)``.
    The count is ``None`` if all warnings of that type should be ignored,
    and the warning type is ``None`` to indicate that the count is about
    all not specified types.

    In case the same type is specified more than once, only the higher
    count is used.
    """
    number_of_warnings = counter.number_of_counts_by(level=level)
    specs = {}
    deduct_all = set()
    for spec_parts in specifications:
        for warning_type, count in spec_parts:
            if count is None:
                deduct_all.add(warning_type)
            else:
                specs[warning_type] = max(specs.get(warning_type, 0), count)
    blanket_ignore = specs.get(None, 0)
    warning_count = counter.counts[level]
    total = number_of_warnings
    for warning_type, count in warning_count.items():
        type_count = warning_count[warning_type]
        if warning_type in specs:
            # Subtract at least 0, and at most the number of warnings counted so
            # the resulting total is guaranteed to be between 0 and `count`.
            total -= max(0, min(count, specs[warning_type]))
        elif warning_type in deduct_all:
            total -= count
        else:
            total -= min(type_count, blanket_ignore)
            blanket_ignore = max(0, blanket_ignore - type_count)
    return total



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a class used to describe a forcefield and all associated data.
"""


import itertools
from glob import glob
import os
from .file_writer import open
from .gmx.rtp import read_rtp
from .ffinput import read_ff
from . import DATA_PATH

FORCE_FIELD_PARSERS = {'.rtp': read_rtp, '.ff': read_ff}

# Cache the force fields.
# It should only be used by the get_native_force_field function, else it would
# allow to request a "native" force field that is not actually native.
_FORCE_FIELDS = {}


class ForceField:
    """
    Description of a force field.

    A force field can be created empty or read from a directory. In any case, a
    force field must be named. If read from a directory, the base name of the
    directory is used as force field name, unless the `name` attribute is
    provided. If the force field is created empty, then `name` must be
    provided.

    Parameters
    ----------
    directory: str or pathlib.Path, optional
        A directory to read the force field from.
    name: str, optional
        The name of the force field.

    Attributes
    ----------
    blocks: dict
    links: list
    modifications: dict
    renamed_residues: dict
    name: str
    variables: dict
    """

    def __init__(self, directory=None, name=None):
        self.blocks = {}
        self.links = []
        self.modifications = {}
        self.renamed_residues = {}
        self.variables = {}
        self.name = None
        if directory is not None:
            self.read_from(directory)
            self.name = os.path.basename(str(directory))
        if name is not None:
            self.name = name
        if self.name is None:
            msg = 'At least one of `directory` or `name` must be provided.'
            raise TypeError(msg)

    def read_from(self, directory):
        """
        Populate or update the force field from a directory.

        The provided directory must contain a subdirectory with the same name
        as the force field.
        """
        source_files = iter_force_field_files(directory)
        for source in source_files:
            extension = os.path.splitext(source)[-1]
            with open(source) as infile:
                FORCE_FIELD_PARSERS[extension](infile, self)

    @property
    def reference_graphs(self):
        """
        Returns all known blocks.

        Returns
        -------
        dict
        """
        return self.blocks

    @property
    def features(self):
        """
        List the features declared by the links.

        Returns
        -------
        set
        """
        return set(feature for link in self.links for feature in link.features)

    def has_feature(self, feature):
        """
        Test if a feature is declared by the links.

        Parameters
        ----------
        feature: str
            The name of the feature of interest.

        Returns
        -------
        bool
        """
        return feature in self.features


def find_force_fields(directory, force_fields=None):
    """
    Read all the force fields in the given directory.

    A force field is defined as a directory that contains at least one RTP
    file. The name of the force field is the base name of the directory.

    If the force field argument is not ``None``, then it must be a dictionary
    with force field names as keys and instances of :class:`ForceField` as
    values. The force fields in the dictionary will be updated if force fields
    with the same names are found in the directory.

    Parameters
    ----------
    directory: pathlib.Path or str
        The path to the directory containing the force fields.
    force_fields: dict
        A dictionary of force fields to update.

    Returns
    -------
    dict
        A dictionary of force fields read or updated. Keys are force field
        names as strings, and values are instances of :class:`ForceField`. If a
        dictionary was provided as the "force_fields" argument, then the
        returned dictionary is the same instance as the one provided but with
        updated content.
    """
    if force_fields is None:
        force_fields = {}
    directory = str(directory)  # Py<3.6 compliance
    for name in os.listdir(directory):
        path = os.path.join(directory, name)
        try:
            next(iter_force_field_files(path))
        except StopIteration:
            pass
        else:
            try:
                if name not in force_fields:
                    force_fields[name] = ForceField(path)
                else:
                    force_fields[name].read_from(path)
            except IOError:
                msg = 'An error occured while reading the force field in  "{}".'
                raise IOError(msg.format(path))
    return force_fields


def iter_force_field_files(directory, extensions=FORCE_FIELD_PARSERS.keys()):
    """
    Returns a generator over the path of all the force field files in the directory.
    """
    return itertools.chain(*(
        glob(os.path.join(str(directory), '*' + extension))
        for extension in extensions
    ))


def get_native_force_field(name):
    """
    Get a force field from the distributed library knowing its name.

    Parameters
    ----------
    name: str
        The name of the requested force field.

    Returns
    -------
    ForceField

    Raises
    ------
    KeyError
        There is no force field with the requested name in the distributed
        library.
    """
    # This function is a *temporary* solution. It reads all the distributed
    # force fields and keep a cache of them. A better solution would only parse
    # the requested force field! There would still be a need to cache the read
    # force fields, though. Indeed, force field comparison is based on instance
    # identity,so we want each force field to be  singleton.
    # TODO: Implement a better way to request a force field by name.
    global _FORCE_FIELDS
    try:
        return _FORCE_FIELDS[name]
    except KeyError:
        _FORCE_FIELDS = find_force_fields(os.path.join(DATA_PATH, 'force_fields'))
        return _FORCE_FIELDS[name]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import combinations
import numbers

import networkx as nx
from numpy import sign

from ..molecule import attributes_match
from .processor import Processor


def _atoms_match(node1, node2):
    return attributes_match(node1, node2, ignore_keys=('order', 'replace'))


def _is_valid_non_edges(molecule, link, rev_raw_match):
    for from_node, to_node_attrs in link.non_edges:
        if from_node not in link:
            continue
        from_mol_node_name = rev_raw_match[from_node]
        from_mol = molecule.nodes[from_mol_node_name]
        # from_link = link.nodes[from_node]
        from_resid = from_mol['resid']
        # from_order = from_link.get('order', 0)
        for neighbor in molecule.neighbors(from_mol_node_name):
            to_mol = molecule.nodes[neighbor]
            to_link = to_node_attrs
            to_resid = to_mol['resid']
            to_order = to_link.get('order', 0)
            if to_resid == from_resid + to_order and _atoms_match(to_mol, to_link):
                return False
    return True


def _pattern_match(molecule, atoms, raw_match):
    for link_key, template_attr in atoms:
        molecule_key = raw_match[link_key]
        molecule_attr = molecule.nodes[molecule_key]
        if not _atoms_match(molecule_attr, template_attr):
            return False
    return True


def _any_pattern_match(molecule, patterns, rev_raw_match):
    return any(_pattern_match(molecule, atoms, rev_raw_match) for atoms in patterns)


def _interpret_order(order):
    error_msg = ('"{}" is not a valid value for the "order" node attribute. '
                 'The value must be an integer, a series of + '
                 '(i.e. >, >>, >>>, ...), a series of <, or a series of *.')
    if order is True or order is False:
        # Booleans match the Number abstract base class, so we have to test
        # for them separately.
        raise ValueError(error_msg.format(order))
    elif isinstance(order, numbers.Number):
        if int(order) != float(order):
            # order is a number but not an int (or int-like)
            raise ValueError(error_msg.format(order))
        order_type = 'number'
        order_value = order
    else:
        try:
            first_character = order[0]
        except (TypeError, IndexError, KeyError):
            # order is not an int, nor a sequence (str, list, tuple,...)
            # or it is an empty sequence. Anyway, we cannot work with it.
            raise ValueError(error_msg.format(order))
        if len(set(order)) != 1 or first_character not in '><*':
            # order is a str (or any sequence, we do not really care),
            # but it contains a mixture of characters (e.g. '+-'), or
            # the characters are not among the ones we expect.
            raise ValueError(error_msg.format(order))
        signs = {'>': +1, '<': -1}
        if first_character in signs:
            order_type = '><'
            order_value = signs[first_character] * len(order)
        elif first_character == '*':
            # This could be an 'else', but it would hide bugs if the code
            # above changes.
            order_type = '*'
            order_value = len(order)
    return order_type, order_value


def match_order(order1, resid1, order2, resid2):
    r"""
    Check if two residues match the order constraints.

    The order can be:

    an integer
        It is then the expected distance in resid with a reference residue.
    a series of >
        This indicates that the residue must have a larger resid than a
        reference residue. Multiple atoms with the same number of > are
        expected to be part of the same residue. The more > are in the serie,
        the further away the residue is expected to be from the reference, so a
        residue with >> is expected to have a greater resid than a residue with
        >.
    a series of <
        Same as a series of >, but for smaller resid.
    a series of *
        This indicates a different residue than the reference, but without a
        specified order. As for the > or the <, atoms with the same number of *
        are expected to be part of the same residue.

    The comparison matrix can be sumerized as follow, with 0 being the
    reference residue, n being an integer. In the matrix, a ? means that the
    result depends on the comparison of the actual numbers, a ! means that the
    comparison should not be considered, and / means that the resids must be
    different. The rows correspond to the order at the left of the comparison
    (order1 argument), while the columns correspond to the order at the right
    of it (order2 argument).

    +-----+---+----+---+----+---+---+----+-----+
    |     | > | >> | < | << | n | 0 | \* | \** |
    +-----+---+----+---+----+---+---+----+-----+
    | >   | = | <  | > | >  | ! | > | !  | !   |
    +-----+---+----+---+----+---+---+----+-----+
    | >>  | > | =  | > | >  | ! | > | !  | !   |
    +-----+---+----+---+----+---+---+----+-----+
    | <   | < | <  | = | >  | ! | < | !  | !   |
    +-----+---+----+---+----+---+---+----+-----+
    | <<  | < | <  | < | =  | ! | < | !  | !   |
    +-----+---+----+---+----+---+---+----+-----+
    | n   | ! | !  | ! | !  | ? | ? | !  | !   |
    +-----+---+----+---+----+---+---+----+-----+
    | 0   | < | <  | > | >  | ? | = | /  | /   |
    +-----+---+----+---+----+---+---+----+-----+
    | \*  | ! | !  | ! | !  | ! | / | =  | /   |
    +-----+---+----+---+----+---+---+----+-----+
    | \** | ! | !  | ! | !  | ! | / | /  | =   |
    +-----+---+----+---+----+---+---+----+-----+

    Parameters
    ----------
    order1: int or str
        The order attribute of the residue on the left of the comparison.
    resid1: int
        The residue id of the residue on the left of the comparison.
    order2: int or str
        The order attribute of the residue on the right of the comparison.
    resid2: int
        The residue id of the residue on the right of the comparison.

    Returns
    -------
    bool
        `True` if the conditions match.

    Raises
    ------
    ValueError
        Raised if the order arguments do not follow the expected format.
    """
    # Validate the order arguments, and format it for what comes next.
    orders = []
    order_types = []
    for order in (order1, order2):
        order_type, order_value = _interpret_order(order)
        order_types.append(order_type)
        orders.append(order_value)

    if order_types[0] == 'number':  # Rows n and 0 in the comparison matrix
        if order_types[1] == 'number':
            # Columns n, and 0
            if (orders[1] - orders[0]) != (resid2 - resid1):
                return False
        elif orders[0] == 0:  # Row 0 in the comparison matrix
            if order_types[1] == '><' and sign(resid2 - resid1) != sign(orders[1]):
                # Columns >, >>, <, and <<
                return False
            elif order_types[1] == '*' and resid1 == resid2:
                # Columns *, and **
                return False
    elif order_types[0] == '><':  # Rows >, >>, <, and <<
        if (order_types[1] == 'number' and orders[1] == 0
                and sign(resid1 - resid2) != sign(orders[0])):
            # Column 0
            return False
        elif (order_types[1] == '><'
              and sign(resid2 - resid1) != sign(orders[1] - orders[0])):
            # Column >, >>, <, and <<
            return False
    elif order_types[0] == '*':  # Rows *, and **
        if order_types[1] == 'number' and orders[1] == 0 and resid1 == resid2:
            # Column 0
            return False
        elif order_types[1] == '*' and ((orders[0] == orders[1]) != (resid1 == resid2)):
            # Columns *, and **
            return False

    return True


def match_link(molecule, link):
    if not attributes_match(molecule.meta, link.molecule_meta):
        return

    GM = nx.isomorphism.GraphMatcher(molecule, link, node_match=_atoms_match)

    raw_matches = GM.subgraph_isomorphisms_iter()
    for raw_match in raw_matches:
        # raw_match: mol -> link
        # rev_raw_match: link -> mol
        rev_raw_match = {value: key for key, value in raw_match.items()}
        if not _is_valid_non_edges(molecule, link, rev_raw_match):
            continue
        any_pattern_match = _any_pattern_match(molecule, link.patterns, rev_raw_match)
        if link.patterns and (not any_pattern_match):
            continue
        order_match = {}
        for mol_idx, link_idx in raw_match.items():
            mol_node = molecule.nodes[mol_idx]
            link_node = link.nodes[link_idx]
            if 'order' in link_node:
                order = link_node['order']
                resid = mol_node['resid']
                if order not in order_match:
                    order_match[order] = resid
                # Assert all orders correspond to the same resid
                elif order in order_match and order_match[order] != resid:
                    break
        else:  # No break
            for ((order1, resid1), (order2, resid2)) in combinations(order_match.items(), 2):
                # Assert the differences between resids correspond to what
                # the orders require.
                if not match_order(order1, resid1, order2, resid2):
                    break
            else:  # No break
                # raw_match is molecule -> link. The other way around is more
                # useful
                yield {v: k for k, v in raw_match.items()}


def _build_link_interaction_from(molecule, interaction, match):
    atoms = tuple(match[idx] for idx in interaction.atoms)
    parameters = [
        param(molecule, match) if callable(param) else param
        for param in interaction.parameters
    ]
    new_interaction = interaction._replace(
        atoms=atoms,
        parameters=parameters
    )
    return new_interaction


class DoLinks(Processor):
    def run_molecule(self, molecule):
        links = molecule.force_field.links
        _nodes_to_remove = []
        for link in links:
            matches = match_link(molecule, link)
            for match in matches:
                for node, node_attrs in link.nodes.items():
                    if 'replace' in node_attrs:
                        if node_attrs['replace'].get('atomname', False) is None:
                            _nodes_to_remove.append(match[node])
                        else:
                            node_mol = molecule.nodes[match[node]]
                            node_mol.update(node_attrs['replace'])
                for inter_type, interactions in link.removed_interactions.items():
                    for interaction in interactions:
                        interaction = _build_link_interaction_from(molecule, interaction, match)
                        try:
                            molecule.remove_matching_interaction(inter_type, interaction)
                        except ValueError:
                            pass
                for inter_type, interactions in link.interactions.items():
                    for interaction in interactions:
                        interaction = _build_link_interaction_from(molecule, interaction, match)
                        molecule.add_or_replace_interaction(inter_type, *interaction)

            molecule.remove_nodes_from(_nodes_to_remove)
        return molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Read force field to force field mappings.
"""

import collections
import itertools
from pathlib import Path

from .file_writer import open
from .log_helpers import StyleAdapter, get_logger
from .map_parser import MappingDirector, Mapping


LOGGER = StyleAdapter(get_logger(__name__))


# TODO: This file contains a parser for backmapping files. This parser needs to
#       be redone using a SectionLineParser like the MappingDirector in
#       map_parser.py

def read_backmapping_file(lines, force_fields):
    """
    Partial reader for modified Backward mapping files.

    Read mappings from a Backward mapping file. Not all fields are supported,
    only the "molecule" and the "atoms" fields are read. If not explicitly
    specified, the origin force field for a molecule is assumed to be
    "universal", and the destination force field is assumed to be "martini22".

    The mapping collection is a 3 level dictionary where the first key is the
    name of the initial force field, the second key is the name of the
    destination force field, and the third key is the name of the molecule.

    Parameters
    ----------
    lines: collections.abc.Iterable[str]
        Collection of lines to read.
    force_fields: dict[str, vermouth.forcefield.ForceField]
        Dict of known force fields.

    Return
    ------
    dict
    """
    lines = iter(lines)
    mappings = collections.defaultdict(lambda: collections.defaultdict(dict))

    # Throw away everything before [ molecule ]
    # If `lines` is empty, then `line_number` is never set by the loop and
    # pylint really does not like it. It should not matter that `line_number`
    # is not set, because it means an exception will be raised anyway. It does
    # not hust to define the variable before the loop, though.
    line_number = None
    for line_number, line in enumerate(lines, start=1):
        cleaned = line.split(';', 1)[0].strip()
        if (cleaned.startswith('[')
                and cleaned.endswith(']')
                and cleaned[1:-1].strip() == 'molecule'):
            break
    else:  # no break
        msg = ('No mapping defined. '
               'A mapping must start with a [ molecule ] section.')
        raise IOError(msg)

    # At this point the last line read was [ molecule ] or else we would have
    # raised an exception.

    # We're going to build a dictionary of {ff_name: {block_name: {atomname: idx}}
    # to make building the mapping objects a little faster.
    name_to_index = collections.defaultdict(dict)
    while True:
        full_mapping = _read_mapping_partial(lines, line_number + 1)
        name, from_ff_list, to_ff_list, mapping, weights, extra, line_number = full_mapping
        if name is None:
            break
        for from_ff, to_ff in itertools.product(from_ff_list, to_ff_list):
            try:
                from_block = force_fields[from_ff].blocks[name]
            except KeyError:
                LOGGER.log(5, "Can't find block {} for FF {}", name, from_ff,
                           type='inconsistent-data')
                continue
            try:
                to_block = force_fields[to_ff].blocks[name]
            except KeyError:
                LOGGER.log(5, "Can't find block {} for FF {}", name, to_ff,
                           type='inconsistent-data')
                continue

            if name not in name_to_index[from_ff]:
                name_to_index[from_ff][name] = _block_names_to_idxs(from_block)
            if name not in name_to_index[to_ff]:
                name_to_index[to_ff][name] = _block_names_to_idxs(to_block)
            map_obj = make_mapping_object(from_block, to_block, mapping,
                                          weights, extra, name_to_index)
            if map_obj is not None:
                mappings[from_ff][to_ff][name] = map_obj

    # We do not want to return defaultdicts.
    mappings = _default_to_dict(mappings)
    return mappings


def _block_names_to_idxs(block):
    """
    Helper function that returns a dict mapping atomnames to node indices for
    every node in block.
    """
    return {block.nodes[idx]['atomname']: idx for idx in block.nodes}


def make_mapping_object(from_block, to_block, mapping, weights, extra, name_to_index):
    """
    Convenience method for creating modern :class:`vermouth.map_parser.Mapping`
    objects from old style mapping information.

    Parameters
    ----------
    from_blocks: collections.abc.Iterable[vermouth.molecule.Block]
    to_blocks: collections.abc.Iterable[vermouth.molecule.Block]
    mapping: dict[tuple[int, str], list[tuple[int, str]]]
        Old style mapping describing what (resid, atomname) maps to what
        (resid, atomname)
    weights: dict[tuple[int, str], dict[tuple[int, str], float]]
        Old style weights, mapping (resid, atomname), (resid, atomname) to a
        weight.
    extra: tuple
    name_to_index: dict[str, dict[str, dict[str, collections.abc.Hashable]]]
        Dict force field names, block names, atomnames to node indices.

    Returns
    -------
    vermouth.map_parser.Mapping
        The created mapping.
    """
    map_dict = collections.defaultdict(lambda: collections.defaultdict(dict))
    from_name_to_idx = name_to_index[from_block.force_field.name][from_block.name]
    to_name_to_idx = name_to_index[to_block.force_field.name][to_block.name]
    for (ridx_from, atname_from), atoms_to in mapping.items():
        for (ridx_to, atname_to) in atoms_to:
            weight = weights[(ridx_to, atname_to)][(ridx_from, atname_from)]
            try:
                idx_from = from_name_to_idx[atname_from]
            except KeyError:
                LOGGER.log(5, "Can't find atom {} in block {} in force field {}",
                           atname_from, from_block.name,
                           from_block.force_field.name,
                           type='inconsistent-data')
                continue
            idx_to = to_name_to_idx[atname_to]
            map_dict[idx_from][idx_to] = weight

    return Mapping(from_block, to_block, map_dict, {},
                   ff_from=from_block.force_field, ff_to=to_block.force_field,
                   extra=extra, normalize_weights=False,
                   type='block', names=(from_block.name,))


def read_mapping_file(lines, force_fields):
    director = MappingDirector(force_fields)
    out = collections.defaultdict(lambda: collections.defaultdict(dict))
    mappings = director.parse(lines)
    for mapping in mappings:
        out[mapping.ff_from][mapping.ff_to][mapping.names] = mapping
    return _default_to_dict(out)


def _default_to_dict(mappings):
    """
    Convert a mapping collection from a defaultdict to a dict.
    """
    if isinstance(mappings, dict):
        return {
            key: _default_to_dict(value)
            for key, value in mappings.items()
        }
    else:
        return mappings


def _compute_weights(mapping, name):
    """
    Calculate the mapping weights from a preliminary mapping.

    The preliminary mapping refers to atoms by their names instead of the tuple
    ``(resid, name)`` used in a final mapping. Target atoms with null weights
    are prefixed with a '!'. The result dictionary also refers to atoms
    directly by name.

    Parameters
    ----------
    mapping: dict[str, list[str]]
        Preliminary mapping.
    name: str
        Molecule name. Used to generate error messages.

    Returns
    -------
    dict
    """
    rev_mapping = collections.defaultdict(list)
    for from_atom, to_atoms in mapping.items():
        for to_atom in to_atoms:
            rev_mapping[to_atom].append(from_atom)

    # Atoms can be mapped with a null weight by prefixing the target particle
    # with a "!". We first set the non-null weights.
    pre_weights = {
        from_atom: dict(collections.Counter([
            to_atom for to_atom in to_atoms if not to_atom.startswith('!')
        ]))
        for from_atom, to_atoms in mapping.items()
    }
    for atom_weights in pre_weights.values():
        total = sum(atom_weights.values())
        for to_atom in atom_weights:
            atom_weights[to_atom] /= total
    weights = collections.defaultdict(dict)
    for from_atom, to_atoms in pre_weights.items():
        for to_atom, weight in to_atoms.items():
            weights[to_atom][from_atom] = weight

    # Then we add the null weights.
    null_weights = {
        to_atom[1:]: {from_atom: 0 for from_atom in from_atoms}
        for to_atom, from_atoms in rev_mapping.items()
        if to_atom.startswith('!')
    }
    for to_atom, from_weights in null_weights.items():
        null_keys = set(from_weights.keys())
        non_null_keys = set(weights.get(to_atom, {}).keys())
        redefined_keys = null_keys & non_null_keys
        if redefined_keys:
            msg = ('Atom(s) {} is mapped to "{}" with and without a weight '
                   'in the molecule "{}". '
                   'There cannot be the same target atom name with and '
                   'without a "!" prefix on a same line.')
            raise IOError(msg.format(redefined_keys, to_atom, name))
        weights[to_atom] = weights.get(to_atom, {})
        weights[to_atom].update(from_weights)

    return weights


def _read_mapping_partial(lines, start_line):
    """
    Partial reader for modified Backward mapping files.

    Read mapping from a Backward mapping file. Not all fields are supported,
    only the "molecule" and the "atoms" fields are read. The origin force field
    is assumed to be "universal", and the destination force field is assumed to
    be "martini22".

    The content is assumed to start just *after* the '[ molecule  ]' line. The
    function consumes the iterator until, and including, the next
    '[ molecule ]' line if any.

    Parameters
    ----------
    lines: collections.abc.Iterator[str]
        Collection of lines to read.
    start_line: int
        The first line number. Starts at 1.

    Returns
    -------
    name: str
        The name of the fragment as read in the "molecule" field.
    from_ff: list[str]
        A list of force field origins. Each force field is referred by name.
    to_ff: list[str]
        A list of force field destinations. Each force field is referred by name.
    mapping: dict[tuple[int, str], list[tuple[int, str]]]
        The mapping. The keys of the dictionary are pairs of residue indices
        and the atom names in the origin force field as tupples (resid,
        atomname); the values are lists of resid, and atom names pairs
        in the destination force field.
    weights: dict
    extra: list
        Unmapped atoms to be added.
    line_number: int
        The number of the last line consumed.
    """
    from_ff = []
    to_ff = []
    mapping = {}
    extra = []
    context = 'molecule'
    name = None

    has_content = False
    line_number = None
    for line_number, line in enumerate(lines, start=start_line):
        cleaned = line.split(';', 1)[0].strip()
        if not cleaned:
            continue
        elif cleaned.startswith('['):
            if not cleaned.endswith(']'):
                raise IOError('Format error at line {}.'.format(line_number))
            context = cleaned[1:-1].strip()
            if context == 'molecule':
                if not has_content:
                    # This detects cases where the file looks like:
                    # [ molecule ]
                    # [ molecule ]
                    #
                    # Then the partial function gets:
                    # [ molecule ]
                    #
                    # as the first section header is consimed by the parent
                    # function.
                    raise IOError('Mapping starting at line {} is empty.'
                                  .format(start_line))
                break
        elif context == 'molecule':
            if name is None:
                name = cleaned
            else:
                msg = ('Line {} tries to redefine the name of the molecule '
                       'starting at line {}.')
                raise IOError(msg.format(line_number, start_line))
        elif context == 'atoms':
            _, from_atom, *to_atoms = cleaned.split()
            if from_atom in mapping:
                msg = ('At line {}, the atom "{}", that is already defined, '
                       'get defined again.')
                raise IOError(msg.format(line_number, from_atom))
            mapping[from_atom] = to_atoms
        elif context in ['from', 'mapping']:
            from_ff.extend(cleaned.split())
        elif context == 'to':
            to_ff.extend(cleaned.split())
        elif context == 'extra':
            extra.extend(cleaned.split())
        # else: we are dealing with a section that we should ignore.
        has_content = True

    if name is None and context != 'molecule':
        # At this point, there are two cases where the name can be None:
        # either it was not defined, or there was no content to read. In the
        # later case, the context was not changed from its initial value.
        # Here, we are in the former case, and the name was not defined. This
        #case is an erroroneous one.
        msg = ('The mapping starting at line {} is defined without a name. '
               'The block name must follow the [ molecule ] section.')
        raise IOError(msg.format(start_line))

    weights = _compute_weights(mapping, name)

    # While it is not supported by the file format, mappings can contain
    # residue information. Atom identifiers in all the outputs must be formated
    # as `(residue, name)`; yet, we only use the name so far. Here we set the
    # residue to 0 for all the atoms assuming we only deal with single residue
    # mappings.
    # We use the opportunity to clean out the null-weight mappings.
    mapping = {
        (0, from_atom): [(0, to_atom if not to_atom.startswith('!') else to_atom[1:])
                         for to_atom in to_atoms]
        for from_atom, to_atoms in mapping.items()
    }
    weights = {
        (0, to_atom): {(0, from_atom): weight
                       for from_atom, weight in from_weights.items()}
        for to_atom, from_weights in weights.items()
    }

    # If not specified in the file, we assume the mapping is for
    # universal -> martini22
    if not from_ff:
        from_ff = ['universal', ]
    if not to_ff:
        to_ff = ['martini22', ]

    return name, from_ff, to_ff, mapping, weights, extra, line_number


def read_mapping_directory(directory, force_fields):
    """
    Read all the mapping files in a directory.

    The resulting mapping collection is a 3-level dict where the keys are:
    * the name of the origin force field
    * the name of the destination force field
    * the name of the residue

    The values after these 3 levels is a mapping dict where the keys are the
    atom names in the origin force field and the values are lists of names in
    the destination force field.

    Parameters
    ----------
    directory: str
        The path to the directory to search. Files with a '.backmap' extension
        will be read. There is no recursive search.
    force_fields: dict[str, ForceField]
        Dict of known forcefields

    Returns
    -------
    dict
        A collection of mappings.
    """
    directory = Path(directory)
    if not directory.is_dir():
        raise NotADirectoryError('"{}" is not a directory.'.format(directory))
    mappings = collections.defaultdict(lambda: collections.defaultdict(dict))
    # Old style mappings
    for path in directory.glob('**/*.map'):
        with open(str(path)) as infile:
            try:
                new_mappings = read_backmapping_file(infile, force_fields)
            except IOError:
                raise IOError('An error occured while reading "{}".'.format(path))
            else:
                combine_mappings(mappings, new_mappings)
    # New style mappings
    for path in directory.glob('**/*.mapping'):
        with open(str(path)) as infile:
            try:
                new_mappings = read_mapping_file(infile, force_fields)
            except IOError:
                raise IOError('An error occured while reading "{}".'.format(path))
            else:
                combine_mappings(mappings, new_mappings)

    return dict(mappings)


def generate_self_mappings(blocks):
    """
    Generate self mappings from a collection of blocks.

    A self mapping is a mapping that maps a force field to itself. Applying
    such mapping is applying a neutral transformation.

    Parameters
    ----------
    blocks: dict[str, networkx.Graph]
        A dictionary of blocks with block names as keys and the blocks
        themselves as values. The blocks must be instances of :class:`networkx.Graph`
        with each node having an 'atomname' attribute.

    Returns
    -------
    mappings: dict[str, tuple]
        A dictionary of mappings where the keys are the names of the blocks,
        and the values are tuples like (mapping, weights, extra).

    Raises
    ------
    KeyError
        Raised if a node does not have am 'atomname' attribute.

    See Also
    --------
    read_mapping_file
        Read a mapping from a file.
    generate_all_self_mappings
        Generate self mappings for a list of force fields.
    """
    mappings = {}
    for name, block in blocks.items():
        mapping = Mapping(block, block, {idx: {idx: 1} for idx in block.nodes},
                          {}, ff_from=block.force_field, ff_to=block.force_field,
                          extra=[], type='block', names=(name,))
        mappings[name] = mapping
    return mappings


def generate_all_self_mappings(force_fields):
    """
    Generate self mappings for a list of force fields.

    Parameters
    ----------
    force_fields: collections.abc.Iterable
        List of instances of :class:`~vermouth.forcefield.ForceField`.

    Returns
    -------
    dict
        A collection of mappings formatted as the output of the
        :func:`read_mapping_directory` function.
    """
    mappings = collections.defaultdict(dict)
    for force_field in force_fields:
        name = force_field.name
        mappings[name][name] = generate_self_mappings(force_field.blocks)
    return dict(mappings)


def combine_mappings(known_mappings, partial_mapping):
    """
    Update a collection of mappings.

    Add the mappings from the 'partial_mapping' argument into the
    'known_mappings' collection. Both arguments are collections of mappings
    similar to the output of the :func:`read_mapping_directory` function. They
    are dictionary with 3 levels of keys: the name of the initial force field,
    the name of the target force field, and the name of the block. The values
    in the third level dictionary are tuples of (mapping, weights, extra).

    If a force field appears in 'partial_mapping' that is not in
    'known_mappings', then it is added. For existing pairs of initial and
    target force fields, the blocks are updated and the version in
    'partial_mapping' is kept in priority.

    Parameters
    ----------
    known_mappings: dict
        Collection of mapping to update **in-place**.
    partial_mapping: dict
        Collection of mappings to update from.
    """
    for origin, destinations in partial_mapping.items():
        known_mappings[origin] = known_mappings.get(origin, {})
        for destination, residues in destinations.items():
            known_mappings[origin][destination] = known_mappings[origin].get(destination, {})
            for residue, mapping in residues.items():
                known_mappings[origin][destination][residue] = mapping



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides processors that can add and remove cystein bridges.
"""

import functools


from ..selectors import proto_multi_templates
from .processor import Processor
from .add_molecule_edges import AddMoleculeEdgesAtDistance
from .. import edge_tuning

UNIVERSAL_BRIDGE_TEMPLATE = [{'resname': 'CYS', 'atomname': 'SG'}, ]


def remove_cystein_bridge_edges(molecule, templates=UNIVERSAL_BRIDGE_TEMPLATE):  # pylint: disable=dangerous-default-value
    """
    Remove all the edges that correspond to cystein bridges from a molecule.

    Cystein bridge edges link an atom from a cystein side chain to the same
    atom on an other cystein. Selecting the correct atom is done with a list
    of template node dictionaries. A template node dictionary functions in the
    same way as node matching in links. An atom that can be involved in a
    cystein bridge must match at least one of the templates of the list.
    The default template list selects the 'SG' bead of the residue 'CYS':
    ``[{'resname': 'CYS', 'atomname': 'SG'}, ]``.

    A template is a dictionary that defines the key:value pairs that must be
    matched in the atoms. Values can be instances of
    :class:`~vermouth.molecule.LinkPredicate`.

    Parameters
    ----------
    molecule: networkx.Graph
        Molecule to modify in-place.
    templates: list[dict]
        A list of templates; selected atom must match at least one.
    """
    selector = functools.partial(proto_multi_templates, templates=templates)
    edge_tuning.prune_edges_with_selectors(molecule, selector)


class RemoveCysteinBridgeEdges(Processor):
    """
    Processor removing edges corresponding to cystein bridges.

    The edge for a cystein bridge is an edge between two atoms that match at
    least one template from a list of templates.

    Parameters
    ----------
    template: list[dict]
        List of node templates.
    """
    def __init__(self, template=UNIVERSAL_BRIDGE_TEMPLATE):  # pylint: disable=dangerous-default-value
        self.template = template

    def run_molecule(self, molecule):
        remove_cystein_bridge_edges(molecule, self.template)
        return molecule


class AddCysteinBridgesThreshold(AddMoleculeEdgesAtDistance):
    """
    Add edges corresponding to cystein bridges on a distance criterion.


    The edge for a cystein bridge is an edge between two atoms that match at
    least one template from a list of templates if the two ends of the edge are
    closer than a given distance.

    Parameters
    ----------
    threshold: float
        Distance in nanometers under which to consider an edge.
    template: list[dict]
        List of node templates.
    """
    def __init__(self, threshold,  # pylint: disable=dangerous-default-value
                 template=UNIVERSAL_BRIDGE_TEMPLATE, attribute='position'):
        super().__init__(
            threshold,
            templates_from=template,
            templates_to=template,
            attribute=attribute,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Merge molecules by chain.
"""

from ..molecule import Molecule
from ..processors.processor import Processor


def merge_chains(system, chains):
    """
    Merge molecules with the given chains as a single molecule.

    Molecules are merged into the resulting molecule if their chain is in the
    list of chains to merge. The resulting molecule is not connected.

    If a molecule comprises multiple chains, then it is merged only if all the
    chains it comprises are part of the selection.

    The meta variable are not conserved in the process.

    The input system is modified in-place.

    Parameters
    ----------
    system: vermouth.system.System
        The system to modify.
    chains: list[str]
        A container of chain identifier.
    """
    chains = set(chains)
    merged = Molecule()
    merged._force_field = system.force_field
    has_merged = False
    new_molecules = []
    for molecule in system.molecules:
        molecule_chains = set(node.get('chain') for node in molecule.nodes.values())
        if molecule_chains.issubset(chains):
            if not has_merged:
                merged.nrexcl = molecule.nrexcl
                new_molecules.append(merged)
            merged.merge_molecule(molecule)
            has_merged = True
        else:
            new_molecules.append(molecule)

    system.molecules = new_molecules


class MergeChains(Processor):
    name = 'MergeChains'

    def __init__(self, chains):
        self.chains = chains

    def run_system(self, system):
        merge_chains(system, self.chains)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that reads a GRO file.

See also
--------
:mod:`vermouth.gmx.gro`
"""


from ..gmx import gro
from .processor import Processor

class GROInput(Processor):
    def __init__(self, filename, exclude=(), ignh=False):
        super().__init__()
        self.filename = filename
        self.exclude = exclude
        self.ignh = ignh

    def run_system(self, system):
        molecule = gro.read_gro(self.filename,
                                exclude=self.exclude,
                                ignh=self.ignh)
        system.add_molecule(molecule)



# Copyright 2019 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the MakeBonds processor.
"""

import pytest
from vermouth.forcefield import get_native_force_field
from vermouth.molecule import Molecule
from vermouth.system import System
from vermouth.processors import MakeBonds


@pytest.mark.parametrize('nodes, edges, expected_edges', (
    [
        # No molecules
        [],
        [],
        []
    ],
    [
        # Single molecule with single node
        [[{'element': 'H', 'position': [0, 0, 0]}], ],
        [[], ],
        [{}, ],
    ],
    [
        # Single molecule with two nodes that should be connected, except
        # they're both hydrogens
        [[{'element': 'H', 'position': [0, 0, 0]},
          {'element': 'H', 'position': [0, 0, 0.12]}], ],
        [[], ],
        [{}],
    ],
    [
        # Two molecule with one node each that should be connected, except
        # they're both hydrogens
        [[{'element': 'H', 'position': [0, 0, 0]}, ],
         [{'element': 'H', 'position': [0, 0, 0.12]}], ],
        [[], []],
        [{}, {}],
    ],
[
        # Single molecule with two nodes that should be connected
        [[{'element': 'C', 'position': [0, 0, 0]},
          {'element': 'H', 'position': [0, 0, 0.12]}], ],
        [[], ],
        [{(0, 1): {'distance': 0.12}}, ],
    ],
    [
        # Two molecule with one node each that should be connected, except that
        # one of the atoms is a H.
        [[{'element': 'C', 'position': [0, 0, 0]}, ],
         [{'element': 'H', 'position': [0, 0, 0.12]}], ],
        [[], []],
        [{}, {}],
    ],
    [
        # Two molecule with one node each that should be connected
        [[{'element': 'C', 'position': [0, 0, 0]}, ],
         [{'element': 'C', 'position': [0, 0, 0.12]}], ],
        [[], []],
        [{(0, 1): {'distance': 0.12}}],
    ],
    [
        # Single molecule with two nodes that should not be connected
        [[{'element': 'H', 'position': [0, 0, 0]},
          {'element': 'H', 'position': [0, 0, 0.2]}], ],
        [[], ],
        [{}],
    ],
    [
        # Two molecule with one node each that should not be connected
        [[{'element': 'H', 'position': [0, 0, 0]}, ],
         [{'element': 'H', 'position': [0, 0, 0.2]}], ],
        [[], []],
        [{}, {}],
    ],
    [
        # Two molecule with one node each that should not be connected because
        # the element is not known
        [[{'element': 'H', 'position': [0, 0, 0]}, ],
         [{'element': 'X', 'position': [0, 0, 0]}], ],
        [[], []],
        [{}, {}],
    ],
    [
        # Single molecule with two nodes that should remain connected
        [[{'element': 'H', 'position': [0, 0, 0]},
          {'element': 'H', 'position': [0, 0, 0.2]}], ],
        [[(0, 1, {'attr': True})], ],
        [{(0, 1): {'attr': True}}],
    ],
    [
        # Single molecule with two nodes that should remain connected
        [[{'element': 'H', 'position': [0, 0, 0]},
          {'element': 'H', 'position': [0, 0, 0.2]}], ],
        [[(0, 1, {})], ],
        [{(0, 1): {}}],
    ],
    [
        # Single molecule with four nodes that should be connected
        [[
            {'element': 'H', 'position': [0, 0, 0]},
            {'element': 'C', 'position': [0, 0, 0.145]},
            {'element': 'C', 'position': [0, 0, 0.315]},
            {'element': 'H', 'position': [0, 0, 0.460]},
        ], ],
        [[], ],
        [{(0, 1): {'distance': pytest.approx(0.145)},
          (1, 2): {'distance': pytest.approx(0.170)},
          (2, 3): {'distance': pytest.approx(0.145)}}],
    ],
    [
        # Single molecule with four nodes that should be connected
        [[
            {'element': 'H', 'position': [0, 0, 0]},
            {'element': 'C', 'position': [0, 0, 0.145]},
            {'element': 'C', 'position': [0, 0, 0.315]},
            {'element': 'H', 'position': [0, 0, 0.460]},
        ], ],
        [[(1, 2, {})], ],
        [{(0, 1): {'distance': pytest.approx(0.145)},
          (1, 2): {},
          (2, 3): {'distance': pytest.approx(0.145)}}],
    ],
    [
        # Single molecule with four nodes that should not be connected despite
        # being close enough because they're not connected in the block
        [[
            {'atomname': 'H1', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0]},
            {'atomname': 'C', 'element': 'C', 'resname': 'GLY', 'position': [0, 0, 0.145]},
            {'atomname': 'N', 'element': 'N', 'resname': 'GLY', 'position': [0, 0, 0.315]},
            {'atomname': 'H2', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0.460]},
        ], ],
        [[(0, 1, {}), (2, 3, {})], ],
        [{(0, 1): {},
          (2, 3): {}}],
    ],
    [
        # Single molecule with four nodes that should be connected despite
        # being far away because they're connected in the block
        [[
            {'atomname': 'H1', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0]},
            {'atomname': 'C', 'element': 'C', 'resname': 'GLY', 'position': [0, 0, 0.145]},
            {'atomname': 'CA', 'element': 'C', 'resname': 'GLY', 'position': [0, 0, 1.315]},
            {'atomname': 'H2', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0.460]},
        ], ],
        [[(0, 1, {}), (2, 3, {})], ],
        [{(0, 1): {},
          (1, 2): {'distance': pytest.approx(1.17)},
          (2, 3): {}}],
    ],
    [
        # Single molecule with four nodes that have a resname, but not atomnames
        [[
            {'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0]},
            {'element': 'C', 'resname': 'GLY', 'position': [0, 0, 0.145]},
            {'element': 'N', 'resname': 'GLY', 'position': [0, 0, 0.315]},
            {'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0.460]},
        ], ],
        [[(0, 1, {}), (2, 3, {})], ],
        [{(0, 1): {},
          (1, 2): {'distance': 0.17},
          (2, 3): {}}],
    ],
    [
        # Single molecule with four nodes that have a resname and duplicate
        # atomnames
        [[
            {'atomname': 'H', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0]},
            {'atomname': 'C', 'element': 'C', 'resname': 'GLY', 'position': [0, 0, 0.145]},
            {'atomname': 'N', 'element': 'N', 'resname': 'GLY', 'position': [0, 0, 0.315]},
            {'atomname': 'H', 'element': 'H', 'resname': 'GLY', 'position': [0, 0, 0.460]},
        ], ],
        [[(0, 1, {}), (2, 3, {})], ],
        [{(0, 1): {},
          (1, 2): {'distance': 0.17},
          (2, 3): {}}],
    ],
))
def test_make_bonds(nodes, edges, expected_edges):
    """
    Test to make sure that make_bonds makes the bonds it is expected to, and not
    too many.
    nodes is a List[List[Dict]], allowing for multiple molecules
    edges is a List[List[Tuple[Int, Int, Dict]]], allowing for
        multiple molecules
    expected_edges is a List[Dict[Tuple[Int, Int], Dict]], allowing for
        multiple molecules
    """
    system = System(force_field=get_native_force_field('universal'))
    for node_set, edge_set in zip(nodes, edges):
        mol = Molecule()
        mol.add_nodes_from(enumerate(node_set))
        mol.add_edges_from(edge_set)
        system.add_molecule(mol)
    MakeBonds().run_system(system)
    # Make sure number of connected components is the same
    assert len(system.molecules) == len(expected_edges)
    # Make sure that for every molecule found, the edges are correct
    for found_mol, ref_edges in zip(system.molecules, expected_edges):
        assert dict(found_mol.edges) == ref_edges



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import os

import os
from setuptools import setup

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

setup(
    package_data={'': package_files('vermouth/data')
                      + package_files('vermouth/tests/data'),},
    scripts=['bin/martinize2', ],
    pbr=True,
)



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the ApplyRubberBand processor and the related functions.
"""
import functools
import pytest
import numpy as np
import networkx as nx
import vermouth
import vermouth.forcefield
from vermouth import selectors
from vermouth.processors import apply_rubber_band
from vermouth.processors.apply_rubber_band import (same_chain,
                                                   are_connected,
                                                   build_connectivity_matrix)

# pylint: disable=redefined-outer-name



@pytest.fixture
def disconnected_graph():
    """
    A graph with two connected components.
    The node keys are integers from 0 to 15 (included). The name of each node
    is its key as a string. The first component is labelled as chain A while
    the second is labelled as chain B.
    0 - 1 - 2 - 3
                |
    4 - 5 - 6 - 7

     8 -  9 - 10 - 11
     |         |    |
    12 - 13 - 14 - 15
    """
    graph = nx.Graph()
    graph.add_nodes_from(range(16))
    graph.add_edges_from([
        # First connected component
        [0, 1], [1, 2], [2, 3], [3, 7], [4, 5], [5, 6], [6, 7],
        # Second connected component
        [8, 9], [8, 12], [8, 9], [9, 10], [10, 14], [10, 11], [11, 15],
        [12, 13], [13, 14], [14, 15],
    ])
    for key, node in graph.nodes.items():
        node['name'] = str(key)
        if key < 8:
            node['chain'] = 'A'
        else:
            node['chain'] = 'B'
    return graph


@pytest.mark.parametrize('selection', (
    list(range(16)),  # Use all the nodes, explicitly
    list(range(8)),  # Only the first component
    list(range(8, 16)),  # Only the second component
    list(range(0, 16, 2)),  # Every other nodes
))
@pytest.mark.parametrize('extra_edges', ([], [(7, 10)]))
def test_build_pair_matrix(disconnected_graph, selection, extra_edges):
    """
    The creation of a pair matrix works as expected.

    Here, the criterion is ``True`` when nodes belong to the same domain.
    The graph is defined as having two domains: one per chain. The extra edges
    allow to make sure the connectivity does not impact the domain detection.
    """
    def have_same_chain(graph, left, right):
        return graph.nodes[left]['chain'] == graph.nodes[right]['chain']

    expected = np.zeros((16, 16), dtype=bool)
    expected[:8, :8] = True
    expected[8:, 8:] = True
    np.fill_diagonal(expected, False)
    expected = expected[:, selection][selection]

    disconnected_graph.add_edges_from(extra_edges)
    idx_to_node = dict(enumerate(disconnected_graph.nodes))
    domains = apply_rubber_band.build_pair_matrix(
        disconnected_graph, have_same_chain, idx_to_node, selection)
    assert np.all(domains == expected)


@pytest.mark.parametrize('separation, outcome', (
    (1, [[0, 2], [2, 6], [4, 6], [8, 12], [8, 10],
         [10, 14], [12, 14]]),
    (2, [[0, 2], [2, 6], [4, 6], [0, 6], [2, 4],
         [8, 12], [8, 10], [10, 14], [12, 14],
         [8, 14], [10, 12]])
))
# this only tests connectivity matrix in terms of connection
# and separation in residue space. selection is tested at the
# end with a more integral test
def test_build_connectivity_matrix(disconnected_graph, separation, outcome):
    idx_to_node = dict(enumerate(disconnected_graph.nodes))
    selection = list(range(0, 16, 2))
    resid_attr = {0: 1, 1: 1, 2: 2, 3: 2, 4: 3, 5: 3, 6: 4, 7: 4,
                  8: 5, 9: 5, 10: 6, 11: 6, 12: 7, 13: 7, 14: 8, 15: 8}
    nx.set_node_attributes(disconnected_graph, resid_attr, 'resid')
    connected = build_connectivity_matrix(disconnected_graph,
                                          separation,
                                          idx_to_node,
                                          selection)
    pairs = []
    for from_idx, to_idx in zip(*np.triu_indices_from(connected)):
        if connected[from_idx, to_idx]:
            idxs = [selection[from_idx], selection[to_idx]]
            idxs.sort()
            pairs.append(idxs)

    for pair in outcome:
        assert pair in pairs

    assert len(pairs) == len(outcome)


@pytest.mark.parametrize('nodes, edges, outcome', (
    ([1, 2, 3],
     [(1, 2), (2, 3)],
     True),
    ([1, 2, 3],
     [(2, 3)],
     False)
))
def test_are_connected(nodes, edges, outcome):
    graph = nx.Graph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    assert are_connected(graph, 1, 2, 1) == outcome


@pytest.mark.parametrize('nodes, chain, edges, outcome', (
    ([1, 2, 3],
     {1: "A", 2: "A", 3: "C"},
     [(1, 2), (2, 3)],
     True),
    ([1, 2, 3],
     {1: "A", 2: "B", 3: "C"},
     [(1, 2), (2, 3)],
     False)
))
def test_same_chain(nodes, edges, chain, outcome):
    graph = nx.Graph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    nx.set_node_attributes(graph, chain, "chain")
    assert same_chain(graph, 1, 2) == outcome


@pytest.fixture
def test_molecule():
    """
    Molecule with the following connectivity and atom-naming:

    SC2:   2           8
           |           |
    SC1:   1   4       7
           |   |       |
    BB:    0 - 3 - 5 - 6
           -------------
    resid: 1   2   3   4  column wise
    """

    force_field = vermouth.forcefield.ForceField("test")
    molecule = vermouth.molecule.Molecule(force_field=force_field)
    molecule.meta['test'] = True
    # The node keys should not be in a sorted order as it would mask any issue
    # due to the keys being accidentally sorted.
    molecule.add_node(2, atomname='SC2',
                      position=np.array([0., 1.0, 0.0]), resid=1)
    molecule.add_node(0, atomname='BB',
                      position=np.array([0., 0., 0.]), resid=1)
    molecule.add_node(1, atomname='SC1',
                      position=np.array([0., 0.5, 0.0]), resid=1)

    molecule.add_node(3, atomname='BB', position=np.array(
        [0.5, 0.0, 0.0]), resid=2)
    molecule.add_node(4, atomname='SC1', position=np.array(
        [0.5, 0.5, 0.0]), resid=2)

    molecule.add_node(5, atomname='BB', position=np.array(
        [1.0, 0.0, 0.0]), resid=3)

    molecule.add_node(6, atomname='BB', position=np.array(
        [1.5, 0.0, 0.0]), resid=4)
    molecule.add_node(7, atomname='SC1', position=np.array(
        [1.5, 0.5, 0.0]), resid=4)
    molecule.add_node(8, atomname='SC2', position=np.array(
        [1.5, 1.0, 0.0]), resid=4)

    molecule.add_edge(0, 1)
    molecule.add_edge(0, 2)
    molecule.add_edge(0, 3)
    molecule.add_edge(3, 4)
    molecule.add_edge(3, 5)
    molecule.add_edge(5, 6)
    molecule.add_edge(6, 7)
    molecule.add_edge(7, 8)

    return molecule


@pytest.mark.parametrize('chain_attribute, atom_names, res_min_dist, outcome',
                         (({0: 'A', 1: 'A', 2: 'A',
                            3: 'A', 4: 'A', 5: 'A',
                            6: 'A', 7: 'A', 8: 'A'},
                           ['BB'],
                           2,
                           [vermouth.molecule.Interaction(
                               atoms=(0, 6),
                               meta={'group': 'Rubber band'},
                               parameters=[6, 1.5, 1000])]
                           ),
                          # different min_res
                          ({0: 'A', 1: 'A', 2: 'A',
                            3: 'A', 4: 'A', 5: 'A',
                            6: 'A', 7: 'A', 8: 'A'},
                           ['BB'],
                           1,
                           [vermouth.molecule.Interaction(
                               atoms=(0, 5),
                               meta={'group': 'Rubber band'},
                               parameters=[6, 1.0, 1000]),
                            vermouth.molecule.Interaction(
                                atoms=(0, 6),
                                meta={'group': 'Rubber band'},
                                parameters=[6, 1.5, 1000]),
                            vermouth.molecule.Interaction(
                                atoms=(3, 6),
                                meta={'group': 'Rubber band'},
                                parameters=[6, 1.0, 1000])]
                           ),
                          # select more than only BB atoms
                          ({0: 'A', 1: 'A', 2: 'A',
                            3: 'A', 4: 'A', 5: 'A',
                            6: 'A', 7: 'A', 8: 'A'},
                           ['BB', 'SC1'],
                           2,
                           [vermouth.molecule.Interaction(
                               atoms=(0, 6),
                               meta={'group': 'Rubber band'},
                               parameters=[6, 1.5, 1000]),
                            vermouth.molecule.Interaction(
                                atoms=(0, 7),
                                meta={'group': 'Rubber band'},
                                parameters=[6, 1.58114, 1000]),
                            vermouth.molecule.Interaction(
                                atoms=(1, 6),
                                meta={'group': 'Rubber band'},
                                parameters=[6, 1.58114, 1000]),
                            vermouth.molecule.Interaction(
                                atoms=(1, 7),
                                meta={'group': 'Rubber band'},
                                parameters=[6, 1.5, 1000])]
                           ),
                          # change chain identifier
                          ({0: 'A', 1: 'A', 2: 'A',
                            3: 'B', 4: 'B', 5: 'B',
                            6: 'B', 7: 'B', 8: 'B'},
                           ['BB'],
                           1,
                           [vermouth.molecule.Interaction(
                               atoms=(3, 6),
                               meta={'group': 'Rubber band'},
                               parameters=[6, 1.0, 1000])]
                           )))
def test_apply_rubber_bands(test_molecule, chain_attribute, atom_names, res_min_dist, outcome):
    """
    Takes molecule and sets the chain attributes. Based on chain, minimum distance
    between residues, and atom names elagible it is tested if rubber bands are applied
    for the correct atoms in molecule.
    """
    selector = functools.partial(
        selectors.proto_select_attribute_in,
        attribute='atomname',
        values=atom_names)

    domain_criterion = vermouth.processors.apply_rubber_band.same_chain
    nx.set_node_attributes(test_molecule, chain_attribute, 'chain')

    process = vermouth.processors.apply_rubber_band.ApplyRubberBand(
        selector=selector,
        lower_bound=0.0,
        upper_bound=10.,
        decay_factor=0,
        decay_power=0.,
        base_constant=1000,
        minimum_force=1,
        bond_type=6,
        domain_criterion=domain_criterion,
        res_min_dist=res_min_dist)
    process.run_molecule(test_molecule)
    assert test_molecule.interactions['bonds'] == outcome



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the writing of ITP file.
"""

# Pylint issues false warnings because of pytest's fixtures.
# pylint: disable=redefined-outer-name
# Some of the expected outputs do contain trailing whitespaces.
# pylint: disable=trailing-whitespace

import io
import textwrap
import pytest
import vermouth
from vermouth.gmx.itp import write_molecule_itp
from vermouth.molecule import Interaction, Molecule


@pytest.fixture
def dummy_molecule():
    """
    A virtually empty molecule.
    """
    molecule = vermouth.Molecule()
    molecule.add_nodes_from((
        (0, {
            'atype': 'A', 'resid': 1, 'resname': 'X', 'atomname': 'A',
            'charge_group': 1, 'charge': 0, 'mass': 72
        }),
        (1, {
            'atype': 'A', 'resid': 1, 'resname': 'X', 'atomname': 'A',
            'charge_group': 1, 'charge': 0, 'mass': 72
        }),
    ))
    molecule.meta['moltype'] = 'TEST'
    molecule.nrexcl = 1
    return molecule


def test_no_header(tmpdir, dummy_molecule):
    """
    Test that no header is written if none is provided.
    """
    outpath = tmpdir / 'out.itp'
    with open(str(outpath), 'w') as outfile:
        write_molecule_itp(dummy_molecule, outfile)

    with open(str(outpath)) as infile:
        assert next(infile) == '[ moleculetype ]\n'


def test_header(tmpdir, dummy_molecule):
    """
    Test that the header is written.
    """
    header = (
        'This is a header.',
        'It contains more than one line.',
    )
    expected = (
        '; This is a header.\n',
        '; It contains more than one line.\n',
        '\n',
    )
    outpath = tmpdir / 'out.itp'
    with open(str(outpath), 'w') as outfile:
        write_molecule_itp(dummy_molecule, outfile, header=header)

    with open(str(outpath)) as infile:
        for line, expected_line in zip(infile, expected):
            assert line == expected_line


@pytest.mark.parametrize('pre_meta', (True, False))
@pytest.mark.parametrize('post_meta', (True, False))
def test_pre_post_section_lines(pre_meta, post_meta):
    """
    :func:`vermouth.gmx.itp.write_molecule_itp` writes pre and post section lines.

    If `pre_meta` or `post_meta` is `True`, then "pre_section_lines" or
    "post_section_lines", respectively, is read from `molecule.meta`, otherwise
    it is passed as an argument.
    """
    molecule = Molecule(nrexcl=1)
    molecule.add_nodes_from((
        (0, {
            'atype': 'Q5', 'resid': 1, 'resname': 'VAL', 'atomname': 'BB',
            'charge_group': 1, 'charge': 1,
        }),
        (1, {
            'atype': 'Q5', 'resid': 1, 'resname': 'VAL', 'atomname': 'BB',
            'charge_group': 1, 'charge': 1,
        }),
        (2, {
            'atype': 'Q5', 'resid': 1, 'resname': 'VAL', 'atomname': 'BB',
            'charge_group': 1, 'charge': 1,
        }),
    ))
    # The molecule has 6 interaction sections. The sections are names
    # "interaction_<joint>_<part>", where <joint> can be "with" or "only", and
    # <part> can be "pre", "post", or "pre_post". The <part> indicates what
    # pre_section_lines and post_section_lines are defined for the section.
    # When <joint> in "with", then the section contains 2 interactions in
    # addition to the pre- and post- lines. When <joint> is "only", then the
    # section only has the pre/post lines defined, and no interactions.
    parts = ('pre', 'post', 'pre_post')
    molecule.interactions = {}
    for part in parts:
        name = 'interaction_with_{}'.format(part)
        molecule.interactions[name] = [
            Interaction(atoms=[0, 1], parameters=[name], meta={}),
            Interaction(atoms=[1, 2], parameters=[name], meta={}),
        ]

    pre_section_lines = {
        'atoms': ['test_atoms_0', 'test_atoms_1'],
        'interaction_with_pre': ['test_pre_0', 'test_pre_1', 'test_pre_2'],
        'interaction_with_pre_post': ['test_pre_post_0'],
        'interaction_only_pre': ['test_pre_only_0', 'test_pre_only_1'],
        'interaction_only_pre_post': ['test_prepost_only_0', 'test_prepost_only_1'],
    }
    post_section_lines = {
        'atoms': ['after_atoms_0'],
        'interaction_with_post': ['after_post_0', 'after_post_1'],
        'interaction_with_pre_post': ['after_pre_post_0'],
        'interaction_only_post': ['after_post_only_0', 'after_post_only_1'],
        'interaction_only_pre_post': ['after_prepost_only_0'],
    }
    expected_segments = [
        """
        [ atoms ]
        test_atoms_0
        test_atoms_1
        1 Q5 1 VAL BB 1 1 
        2 Q5 1 VAL BB 1 1 
        3 Q5 1 VAL BB 1 1 
        after_atoms_0
        """,
        """
        [ interaction_with_pre ]
        test_pre_0
        test_pre_1
        test_pre_2
        1 2 interaction_with_pre
        2 3 interaction_with_pre
        """,
        """
        [ interaction_with_post ]
        1 2 interaction_with_post
        2 3 interaction_with_post
        """,
        """
        [ interaction_with_pre_post ]
        test_pre_post_0
        1 2 interaction_with_pre_post
        2 3 interaction_with_pre_post
        after_pre_post_0
        """,
        """
        [ interaction_only_pre ]
        test_pre_only_0
        test_pre_only_1
        """,
        """
        [ interaction_only_post ]
        after_post_only_0
        after_post_only_1
        """,
        """
        [ interaction_only_pre_post ]
        test_prepost_only_0
        test_prepost_only_1
        after_prepost_only_0
        """,
    ]

    if post_meta:
        arg_post = post_section_lines
        molecule.meta['post_section_lines'] = 'invalid'
    else:
        arg_post = None
        molecule.meta['post_section_lines'] = post_section_lines
    if pre_meta:
        arg_pre = pre_section_lines
        molecule.meta['pre_section_lines'] = 'invalid'
    else:
        arg_pre = None
        molecule.meta['pre_section_lines'] = pre_section_lines

    outfile = io.StringIO()
    write_molecule_itp(molecule, outfile, moltype='test',
                       post_section_lines=arg_post, pre_section_lines=arg_pre)
    itp_content = outfile.getvalue()
    # This could be a assert all(...), but it makes it more difficult to
    # understant what is happening in case of a failure.
    for segment in expected_segments:
        assert textwrap.dedent(segment)[:-1] in itp_content



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test graph reparation and related operations.
"""
import copy
import logging

import pytest
import vermouth
from vermouth.molecule import Link
import vermouth.forcefield

# pylint: disable=redefined-outer-name

def build_forcefield_with_mods():
    """
    Build a force field that describe some modifications.

    Returns
    -------
    vermouth.ForceField
    """
    nter = Link(name='N-ter')
    nter.add_nodes_from((
        (0, {'atomname': 'CA', 'PTM_atom': False, 'element': 'C'}),
        (1, {'atomname': 'N', 'PTM_atom': False, 'element': 'N'}),
        (2, {'atomname': 'HN', 'PTM_atom': False, 'element': 'H'}),
        (3, {'atomname': 'H', 'PTM_atom': True, 'element': 'H'}),
    ))
    nter.add_edges_from([[0, 1], [1, 2], [1, 3]])

    cter = Link(name='C-ter')
    cter.add_nodes_from((
        (0, {'atomname': 'C', 'PTM_atom': False, 'element': 'C'}),
        (1, {'atomname': 'O', 'PTM_atom': False, 'element': 'O'}),
        (2, {'atomname': 'OXT', 'PTM_atom': True, 'element': 'O'}),
    ))
    cter.add_edges_from([[0, 1], [0, 2]])

    gluh = Link(name='GLU-H')
    gluh.add_nodes_from((
        (0, {'atomname': 'CD', 'PTM_atom': False, 'element': 'C'}),
        (1, {'atomname': 'OE1', 'PTM_atom': False, 'element': 'O'}),
        (2, {'atomname': 'OE2', 'PTM_atom': False, 'element': 'O'}),
        (3, {'atomname': 'HE1', 'PTM_atom': True, 'element': 'H'}),
    ))
    gluh.add_edges_from([[0, 1], [0, 2], [1, 3]])

    forcefield = copy.copy(vermouth.forcefield.get_native_force_field('universal'))
    forcefield.modifications = {}
    for mod in [nter, gluh, cter]:
        forcefield.modifications[mod.name] = mod
    forcefield.renamed_residues[('GLU', ('GLU-H', 'N-ter'))] = 'GLU0'
    return forcefield


def build_system_mod(force_field):
    """
    Build a system with a molecule that comprises modifications.

    Parameters
    ----------
    force_field: vermouth.ForceField
        A force field based on "universal" that desribed the "GLU-H", "C-ter"
        and "N-ter" modifications.

    Returns
    -------
    vermouth.System

    See Also
    --------
    build_forcefield_with_mods
    """
    molecule = vermouth.molecule.Molecule()
    # When the name must be fixed, the expected name is stored under 'expected'.
    # If more than one value is acceptable, then 'expected' is a tuple.
    molecule.add_nodes_from((
        # Residue 1 is a N-terminus (atom 14, H) and has a protonated
        # side chain (atom 13, HE1)
        (0, {'resid': 1, 'resname': 'GLU', 'atomname': 'N', 'chain': 'A', 'element': 'N'}),
        (1, {'resid': 1, 'resname': 'GLU', 'atomname': 'CA', 'chain': 'A', 'element': 'C'}),
        (2, {'resid': 1, 'resname': 'GLU', 'atomname': 'C', 'chain': 'A', 'element': 'C'}),
        (3, {'resid': 1, 'resname': 'GLU', 'atomname': 'O', 'chain': 'A', 'element': 'O'}),
        (4, {'resid': 1, 'resname': 'GLU', 'atomname': 'CB', 'chain': 'A', 'element': 'C'}),
        (5, {'resid': 1, 'resname': 'GLU', 'atomname': 'HB1', 'chain': 'A', 'element': 'H'}),
        (6, {'resid': 1, 'resname': 'GLU', 'atomname': 'HB2', 'chain': 'A', 'element': 'H'}),
        (7, {'resid': 1, 'resname': 'GLU', 'atomname': 'CG', 'chain': 'A', 'element': 'C'}),
        (8, {'resid': 1, 'resname': 'GLU', 'atomname': 'HG1', 'chain': 'A', 'element': 'H'}),
        (9, {'resid': 1, 'resname': 'GLU', 'atomname': 'HG2', 'chain': 'A', 'element': 'H'}),
        (10, {'resid': 1, 'resname': 'GLU', 'atomname': 'CD', 'chain': 'A', 'element': 'C'}),
        (11, {'resid': 1, 'resname': 'GLU', 'atomname': 'OE2', 'chain': 'A', 'element': 'O'}),
        (12, {'resid': 1, 'resname': 'GLU', 'atomname': 'OE1', 'chain': 'A', 'element': 'O'}),
        (13, {'resid': 1, 'resname': 'GLU', 'atomname': 'HE1', 'chain': 'A', 'element': 'H'}),
        (14, {'resid': 1, 'resname': 'GLU', 'atomname': 'H', 'chain': 'A', 'element': 'H'}),
        (15, {'resid': 1, 'resname': 'GLU', 'atomname': 'HA', 'chain': 'A', 'element': 'H'}),
        (16, {'resid': 1, 'resname': 'GLU', 'atomname': 'HN', 'chain': 'A', 'element': 'H'}),
        # Residue 2 has missing atoms (HA1 and HA2, the hydrogen of the C
        # alpha; HN the hydrogen on the N-ter nitrogen and O, the oxygen at the
        # C-ter). 'common' is an arbitrary attribute name with identical values,
        # which should be propagated to reconstructed atoms.
        (17, {'resid': 2, 'resname': 'GLY', 'atomname': 'N', 'chain': 'A', 'element': 'N', 'common': 'a'}),
        (18, {'resid': 2, 'resname': 'GLY', 'atomname': 'CA', 'chain': 'A', 'element': 'C', 'common': 'a'}),
        (19, {'resid': 2, 'resname': 'GLY', 'atomname': 'C', 'chain': 'A', 'element': 'C', 'common': 'a'}),
        # Residue 3 has only N and C specified, everything else is missing.
        (20, {'resid': 3, 'resname': 'GLY', 'atomname': 'N', 'chain': 'A', 'element': 'N'}),
        (21, {'resid': 3, 'resname': 'GLY', 'atomname': 'O', 'chain': 'A', 'element': 'O'}),
        # Residue 4 has shuffled and missing names
        (22, {'resid': 4, 'resname': 'GLY', 'atomname': 'CA', 'chain': 'A', 'element': 'N', 'expected': 'N'}),
        (23, {'resid': 4, 'resname': 'GLY', 'chain': 'A', 'element': 'C', 'expected': 'CA'}),
        (24, {'resid': 4, 'resname': 'GLY', 'atomname': 'HA1', 'chain': 'A', 'element': 'C', 'expected': 'C'}),
        (25, {'resid': 4, 'resname': 'GLY', 'atomname': 'HN', 'chain': 'A', 'element': 'O', 'expected': 'O'}),
        (26, {'resid': 4, 'resname': 'GLY', 'atomname': 'N', 'chain': 'A', 'element': 'H', 'expected': ('HA1', 'HA2')}),
        (27, {'resid': 4, 'resname': 'GLY', 'atomname': 'N', 'chain': 'A', 'element': 'H', 'expected': ('HA1', 'HA2')}),
        (28, {'resid': 4, 'resname': 'GLY', 'atomname': '', 'chain': 'A', 'element': 'H', 'expected': 'HN'}),
        # Residue 5 has a wrong name in a modification, and a wrong name for
        # the anchor.
        (29, {'resid': 5, 'resname': 'GLY', 'atomname': 'N', 'chain': 'A', 'element': 'N'}),
        (30, {'resid': 5, 'resname': 'GLY', 'atomname': 'CA', 'chain': 'A', 'element': 'C'}),
        # Should be C
        (31, {'resid': 5, 'resname': 'GLY', 'atomname': 'WRONG', 'chain': 'A', 'element': 'C', 'expected': 'C'}),
        (32, {'resid': 5, 'resname': 'GLY', 'atomname': 'O', 'chain': 'A', 'element': 'O'}),
        (33, {'resid': 5, 'resname': 'GLY', 'atomname': 'HN', 'chain': 'A', 'element': 'H'}),
        (34, {'resid': 5, 'resname': 'GLY', 'atomname': 'HA1', 'chain': 'A', 'element': 'H'}),
        (35, {'resid': 5, 'resname': 'GLY', 'atomname': 'HA2', 'chain': 'A', 'element': 'H'}),
        # Should be OXT
        (36, {'resid': 5, 'resname': 'GLY', 'atomname': 'O2', 'chain': 'A', 'element': 'O', 'expected': 'OXT'}),
    ))
    molecule.add_edges_from([
        [0, 1], [1, 2], [2, 3], [1, 4], [4, 5], [4, 6], [4, 7], [7, 8],
        [7, 9], [7, 10], [10, 11], [10, 12], [12, 13], [14, 0], [15, 1], [16, 0],
        [2, 17], [17, 18], [18, 19],
        [19, 20], [21, 22], [22, 23], [23, 24], [24, 25],
        [22, 28], [23, 26], [23, 27], [24, 29], [29, 30], [30, 31], [31, 32],
        [31, 36], [29, 33], [30, 34], [30, 35],
    ])
    for node in molecule:
        molecule.nodes[node]['atomid'] = node
    system = vermouth.System()
    system.molecules = [molecule]
    system.force_field = force_field
    return system


@pytest.fixture
def forcefield_with_mods():
    return build_forcefield_with_mods()


@pytest.fixture
def system_mod(forcefield_with_mods):
    return build_system_mod(forcefield_with_mods)


@pytest.fixture(params=(True, False))
def repaired_graph(request, system_mod):
    vermouth.RepairGraph(include_graph=request.param).run_system(system_mod)
    return system_mod


@pytest.fixture
def canonicalized_graph(repaired_graph):
    vermouth.CanonicalizeModifications().run_system(repaired_graph)
    return repaired_graph


@pytest.fixture
def renamed_graph(canonicalized_graph):
    vermouth.RenameModifiedResidues().run_system(canonicalized_graph)
    return canonicalized_graph


@pytest.mark.parametrize('node_key', (13, 14, 36))
def test_PTM_atom_true(repaired_graph, node_key):
    molecule = repaired_graph.molecules[0]
    assert molecule.nodes[node_key].get('PTM_atom', False)


@pytest.mark.parametrize('node_key', (
    node_key for node_key in range(16) if node_key not in (13, 14)
))
def test_PTM_atom_false(repaired_graph, node_key):
    molecule = repaired_graph.molecules[0]
    assert not molecule.nodes[node_key].get('PTM_atom', False)


def test_uniq_names_repaired(repaired_graph):
    assert len(repaired_graph.molecules) == 1
    molecule = repaired_graph.molecules[0]
    # Only for non-PTM atoms, since those names are not touched by repair graph
    atoms = [(node['resid'], node['atomname'])
             for node in molecule.nodes.values()
             if not node.get('PTM_atom', False)]
    deduplicated = set(atoms)
    assert len(atoms) == len(deduplicated)


def test_uniq_names_canonicalize(canonicalized_graph):
    assert len(canonicalized_graph.molecules) == 1
    molecule = canonicalized_graph.molecules[0]
    atoms = [(node['resid'], node['atomname'])
             for node in molecule.nodes.values()]
    deduplicated = set(atoms)
    assert len(atoms) == len(deduplicated)


def _list_expected_names():
    molecule = build_system_mod(build_forcefield_with_mods()).molecules[0]
    for key, node in molecule.nodes.items():
        if 'expected' in node:
            expected = node['expected']
        else:
            expected = node['atomname']
        if not isinstance(expected, tuple):
            expected = (expected, )
        yield key, expected


@pytest.mark.parametrize('key, expected_names', _list_expected_names())
def test_name_repaired(repaired_graph, key, expected_names):
    molecule = repaired_graph.molecules[0]
    if not molecule.nodes[key].get('PTM_atom', False):
        print(key, molecule.nodes[key])
        assert molecule.nodes[key]['atomname'] in expected_names


@pytest.mark.parametrize('key, expected_names', _list_expected_names())
def test_name_canonicalized(canonicalized_graph, key, expected_names):
    molecule = canonicalized_graph.molecules[0]
    assert molecule.nodes[key]['atomname'] in expected_names


@pytest.mark.parametrize('resid, expected_attrs', [
    (2, {'resid': 2, 'resname': 'GLY', 'chain': 'A', 'common': 'a'})
])
def test_common_attributes(repaired_graph, resid, expected_attrs):
    """Test that attributes that are common to all nodes in a residue end up in
    reconstructed nodes."""
    mol = repaired_graph.molecules[0]
    for node_idx in mol:
        node = mol.nodes[node_idx]
        if node.get('resid') == resid:
            for key, val in expected_attrs.items():
                assert key in node
                assert node[key] == val


def test_renaming(renamed_graph):
    for node in renamed_graph.molecules[0].nodes.values():
        if node['resid'] == 1:
            assert node['resname'] == 'GLU0'
        else:
            assert node['resname'] == 'GLY'



@pytest.mark.parametrize('resid,mutations,modifications,atomnames', [
    (1, ['ALA'], [], 'O C CA HA N HN CB HB1 HB2 HB3'),  # The glutamate chain and N-ter are removed
    (1, [], ['N-ter'], 'O C CA HA N H HN CB HB1 HB2 CG HG1 HG2 CD OE1 OE2'),  # HE1 got removed
    (2, ['ALA'], ['N-ter', 'C-ter'], 'O OXT C CA HA N H HN CB HB1 HB2 HB3'),
    (2, ['GLU'], [], 'O C CA HA N HN CB HB1 HB2 CG HG1 HG2 CD OE1 OE2'),  # Added glutamate sidechain
    (5, ['GLY'], ['none'], 'N CA C O HN HA1 HA2'),  # Remove O2 from C-ter mod
])
def test_repair_graph_with_mutation_modification(system_mod, resid, mutations,
                                                 modifications, atomnames):
    mol = system_mod.molecules[0]
    # Let's mutate res1 to ALA
    for node_idx in mol:
        if mol.nodes[node_idx].get('resid') == resid:
            if mutations:
                mol.nodes[node_idx]['mutation'] = mutations
            if modifications:
                mol.nodes[node_idx]['modification'] = modifications
    mol = vermouth.RepairGraph().run_molecule(mol)
    resid1_atomnames = set()
    for node_idx in mol:
        if mol.nodes[node_idx].get('resid') == resid:
            if mutations:
                assert mol.nodes[node_idx]['resname'] == mutations[0]
            if modifications:
                assert mol.nodes[node_idx].get('modification') == modifications
            resid1_atomnames.add(mol.nodes[node_idx]['atomname'])
    assert resid1_atomnames == set(atomnames.split())


@pytest.mark.parametrize('resid,mutations,modifications', [
    (2, [], ['GLU-H']),  # The glutamate chain and N-ter are removed
    (2, ['ALA', 'LEU'], [])
])
def test_repair_graph_with_mutation_modification_error(system_mod, caplog,
                                                       resid, mutations,
                                                       modifications):
    mol = system_mod.molecules[0]
    # Let's mutate res1 to ALA
    for node_idx in mol:
        if mol.nodes[node_idx].get('resid') == resid:
            if mutations:
                mol.nodes[node_idx]['mutation'] = mutations
            if modifications:
                mol.nodes[node_idx]['modification'] = modifications
    with pytest.raises(ValueError), caplog.at_level(logging.WARNING):
        assert not caplog.records
        mol = vermouth.RepairGraph().run_molecule(mol)
        assert len(caplog.records) == 1


@pytest.mark.parametrize('known_mod_names', [
    [],
    ['C-ter'],
    ['C-ter', 'N-ter'],
    ['GLU-H', 'N-ter'],
])
def test_unknown_mods_removed(caplog, repaired_graph, known_mod_names):
    """
    Tests that atoms that are part of modifications, but are not recognized, get
    removed from the graph by CanonicalizeModifications
    """
    caplog.set_level(logging.WARNING)
    ff = copy.copy(repaired_graph.force_field)
    for mod_name in known_mod_names:
        assert mod_name in ff.modifications  # Purely defensive

    removed_mods = []
    for name, mod in dict(ff.modifications).items():
        if name not in known_mod_names:
            del ff.modifications[name]
            removed_mods.append(mod)

    repaired_graph.force_field = ff
    mol = repaired_graph.molecules[0]

    assert not caplog.records
    assert len(mol) == 46
    vermouth.CanonicalizeModifications().run_system(repaired_graph)

    assert caplog.records

    for record in caplog.records:
        assert record.levelname == 'WARNING'

    assert len(mol) < 46
    atomnames = dict(mol.nodes(data='atomname')).values()
    for mod in removed_mods:
        for node_key in mod.nodes:
            node = mod.nodes[node_key]
            if node['PTM_atom']:
                assert node['atomname'] not in atomnames

    for node_key in mol.nodes:
        node = mol.nodes[node_key]
        if node.get('PTM_atom'):
            contained_by = [mod for mod in ff.modifications.values()
                            if node.get('expected', node['atomname']) in
                               dict(mod.nodes(data='atomname')).values()]
            assert len(contained_by) == 1



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a processor that merges all the molecules from a system.
"""

from .processor import Processor


class MergeAllMolecules(Processor):
    """
    Merge all the molecules from a system.

    The molecules are merged into the first molecule of the system. Nothing is
    done if there are no molecules.
    """
    def run_system(self, system):
        if not system.molecules:
            return system

        molecule = system.molecules[0]
        for other in system.molecules[1:]:
            molecule.merge_molecule(other)

        system.molecules = [molecule]
        return system

    @staticmethod
    def run_molecule(molecule):
        raise NotImplementedError('MergeAllMolecules only works on systems.')
